#!/usr/bin/env python3
import json
import os
import threading
import time
from pathlib import Path

from PyQt5.QtCore import QEvent, QObject, QTimer
from PyQt5.QtWidgets import QAction, QDialog, QDockWidget

from classes import info
from classes.app import get_app
from classes.logger import log


def _env_true(name, default=False):
    value = os.getenv(name)
    if value is None:
        return default
    return str(value).strip().lower() in ("1", "true", "yes", "on")


def _env_float(name, default):
    value = os.getenv(name)
    if value is None:
        return float(default)
    try:
        return float(value)
    except Exception:
        return float(default)


class UiTraceRecorder(QObject):
    """Record UI traces in a centralized, env-configured subsystem."""

    def __init__(self, window):
        super().__init__(window)
        self.window = window
        self._lock = threading.Lock()
        self._seq = 0
        self._closed = False
        self._updates_handle = None
        self._selections_handle = None
        self._events_handle = None
        self._dialog_connections = {}
        self._action_connections = []
        self._dock_connections = []
        self._cache_timer = None
        self._last_cache_signature = None
        self._updates_listener_connected = False
        self._selection_connected = False
        self._dialog_filter_installed = False

        self._include_load = _env_true("OPENSHOT_UI_TRACE_INCLUDE_LOAD", default=False)
        self._include_ignored = _env_true("OPENSHOT_UI_TRACE_INCLUDE_IGNORED", default=False)
        self._trace_actions = _env_true("OPENSHOT_UI_TRACE_ACTIONS", default=True)
        self._trace_dialogs = _env_true("OPENSHOT_UI_TRACE_DIALOGS", default=True)
        self._trace_docks = _env_true("OPENSHOT_UI_TRACE_DOCKS", default=True)
        self._trace_cache = _env_true("OPENSHOT_UI_TRACE_CACHE", default=False)
        self._cache_interval_ms = max(100, int(_env_float("OPENSHOT_UI_TRACE_CACHE_INTERVAL", 0.5) * 1000))
        self.enabled = False

        updates_path, selections_path, events_path = self._resolve_paths()
        if not updates_path and not selections_path and not events_path:
            return

        try:
            if updates_path:
                self._updates_handle = self._open_jsonl(updates_path)
            if selections_path:
                self._selections_handle = self._open_jsonl(selections_path)
            if events_path:
                self._events_handle = self._open_jsonl(events_path)

            if self._updates_handle or self._events_handle:
                get_app().updates.add_listener(self)
                self._updates_listener_connected = True
            if self._selections_handle or self._events_handle:
                self.window.SelectionChanged.connect(self._on_selection_changed)
                self._selection_connected = True
            if self._events_handle and self._trace_actions:
                self._connect_actions()
            if self._events_handle and self._trace_dialogs:
                get_app().installEventFilter(self)
                self._dialog_filter_installed = True
            if self._events_handle and self._trace_docks:
                self._connect_docks()
            if self._events_handle and self._trace_cache:
                self._start_cache_timer()

            self.enabled = True
            self._write_meta()
            log.info(
                "UI trace recorder enabled. updates=%s selections=%s events=%s",
                updates_path,
                selections_path,
                events_path,
            )
        except Exception:
            log.error("Failed to initialize UI trace recorder", exc_info=1)
            self.close()

    def _resolve_paths(self):
        updates_path = os.getenv("OPENSHOT_UI_TRACE_UPDATES", "").strip()
        selections_path = os.getenv("OPENSHOT_UI_TRACE_SELECTIONS", "").strip()
        events_path = os.getenv("OPENSHOT_UI_TRACE_EVENTS", "").strip()

        trace_dir = os.getenv("OPENSHOT_UI_TRACE_DIR", "").strip()
        trace_enabled = _env_true("OPENSHOT_UI_TRACE", default=False)

        if trace_dir and (trace_enabled or updates_path or selections_path or events_path):
            stamp = time.strftime("%Y%m%d_%H%M%S")
            pid = os.getpid()
            trace_root = Path(trace_dir).expanduser().resolve()
            trace_root.mkdir(parents=True, exist_ok=True)
            if not updates_path:
                updates_path = str(trace_root / f"updates_{stamp}_{pid}.jsonl")
            if not selections_path:
                selections_path = str(trace_root / f"selections_{stamp}_{pid}.jsonl")
            if not events_path:
                events_path = str(trace_root / f"events_{stamp}_{pid}.jsonl")

        if not trace_enabled and not updates_path and not selections_path and not events_path:
            return None, None, None

        return updates_path or None, selections_path or None, events_path or None

    @staticmethod
    def _open_jsonl(path):
        p = Path(path).expanduser().resolve()
        p.parent.mkdir(parents=True, exist_ok=True)
        return open(p, "a", encoding="utf-8", buffering=1)

    def _next_seq(self):
        with self._lock:
            self._seq += 1
            return self._seq

    def _event_base(self, kind):
        return {
            "event": kind,
            "seq": self._next_seq(),
            "ts": round(time.time(), 6),
        }

    def _write_jsonl(self, handle, payload):
        if handle is None or self._closed:
            return
        line = json.dumps(payload, ensure_ascii=False, default=str)
        handle.write(line + "\n")

    def _write_event(self, payload, include_updates=False, include_selections=False):
        self._write_jsonl(self._events_handle, payload)
        if include_updates:
            self._write_jsonl(self._updates_handle, payload)
        if include_selections:
            self._write_jsonl(self._selections_handle, payload)

    def _write_meta(self):
        payload = self._event_base("meta")
        payload.update(
            {
                "pid": os.getpid(),
                "include_load": self._include_load,
                "include_ignored": self._include_ignored,
                "trace_actions": self._trace_actions,
                "trace_dialogs": self._trace_dialogs,
                "trace_docks": self._trace_docks,
                "trace_cache": self._trace_cache,
                "cache_interval_ms": self._cache_interval_ms,
                "project_path": str(get_app().project.current_filepath or ""),
            }
        )
        self._write_event(payload, include_updates=True, include_selections=True)

    def changed(self, action):
        if (self._updates_handle is None and self._events_handle is None) or self._closed:
            return
        try:
            ignore_history = bool(get_app().updates.ignore_history)
            if action.type == "load" and not self._include_load:
                return
            if ignore_history and not self._include_ignored:
                return
            payload = self._event_base("update")
            payload.update(
                {
                    "action_type": action.type,
                    "transaction": action.transaction,
                    "key": action.key,
                    "value": action.values,
                    "old_values": action.old_values,
                    "ignore_history": ignore_history,
                    "data_version": int(get_app().updates.data_version),
                }
            )
            self._write_event(payload, include_updates=True)
        except Exception:
            log.error("UI trace update write failed", exc_info=1)

    def _on_selection_changed(self):
        if (self._selections_handle is None and self._events_handle is None) or self._closed:
            return
        try:
            payload = self._event_base("selection")
            payload.update(
                {
                    "selected_items": json.loads(json.dumps(self.window.selected_items)),
                    "selected_clips": list(self.window.selected_clips),
                    "selected_transitions": list(self.window.selected_transitions),
                    "selected_effects": list(self.window.selected_effects),
                    "selected_tracks": list(self.window.selected_tracks),
                    "selected_markers": list(self.window.selected_markers),
                    "show_property_id": self.window.show_property_id,
                    "show_property_type": self.window.show_property_type,
                }
            )
            self._write_event(payload, include_selections=True)
        except Exception:
            log.error("UI trace selection write failed", exc_info=1)

    def _connect_actions(self):
        for action in self.window.findChildren(QAction):
            try:
                slot = self._make_action_slot(action)
                action.triggered.connect(slot)
                self._action_connections.append((action, slot))
            except Exception:
                log.debug("Failed to connect QAction trace for %s", action.objectName(), exc_info=1)

    def _make_action_slot(self, action):
        def _slot(checked=False):
            self._on_action_triggered(action, checked)

        return _slot

    def _on_action_triggered(self, action, checked):
        if self._events_handle is None or self._closed:
            return
        payload = self._event_base("action_triggered")
        payload.update(
            {
                "action_name": action.objectName() or "",
                "action_text": (action.text() or "").replace("&", ""),
                "checked": bool(checked),
            }
        )
        self._write_event(payload)

    def _connect_docks(self):
        for dock in self.window.findChildren(QDockWidget):
            try:
                visibility_slot = self._make_dock_visibility_slot(dock)
                dock.visibilityChanged.connect(visibility_slot)
                self._dock_connections.append((dock.visibilityChanged, visibility_slot))
            except Exception:
                log.debug("Failed to connect dock visibility trace for %s", dock.objectName(), exc_info=1)

    def _make_dock_visibility_slot(self, dock):
        def _slot(visible):
            self._on_dock_visibility_changed(dock, visible)

        return _slot

    def _on_dock_visibility_changed(self, dock, visible):
        if self._events_handle is None or self._closed:
            return
        payload = self._event_base("dock_visibility")
        payload.update(
            {
                "object_name": dock.objectName() or "",
                "window_title": dock.windowTitle() or "",
                "visible": bool(visible),
                "floating": bool(dock.isFloating()),
            }
        )
        self._write_event(payload)

    def eventFilter(self, obj, event):
        if self._closed or not self._trace_dialogs or self._events_handle is None:
            return False
        try:
            if isinstance(obj, QDialog):
                event_type = event.type()
                if event_type == QEvent.Show:
                    self._attach_dialog(obj)
                    self._emit_dialog("shown", obj)
                elif event_type == QEvent.Hide:
                    self._emit_dialog("hidden", obj)
                elif event_type == QEvent.Close:
                    self._emit_dialog("closed", obj)
        except Exception:
            log.debug("UI trace dialog eventFilter failed", exc_info=1)
        return False

    def _attach_dialog(self, dialog):
        key = id(dialog)
        if key in self._dialog_connections:
            return
        slots = []
        try:
            accepted_slot = lambda: self._emit_dialog("accepted", dialog, result="accepted")
            dialog.accepted.connect(accepted_slot)
            slots.append((dialog.accepted, accepted_slot))
        except Exception:
            pass
        try:
            rejected_slot = lambda: self._emit_dialog("rejected", dialog, result="rejected")
            dialog.rejected.connect(rejected_slot)
            slots.append((dialog.rejected, rejected_slot))
        except Exception:
            pass
        try:
            finished_slot = lambda code: self._emit_dialog("finished", dialog, result=int(code))
            dialog.finished.connect(finished_slot)
            slots.append((dialog.finished, finished_slot))
        except Exception:
            pass
        self._dialog_connections[key] = slots

    def _emit_dialog(self, phase, dialog, result=None):
        payload = self._event_base("dialog_lifecycle")
        payload.update(
            {
                "phase": phase,
                "class_name": dialog.__class__.__name__,
                "object_name": dialog.objectName() or "",
                "window_title": dialog.windowTitle() or "",
                "modal": bool(dialog.isModal()),
            }
        )
        if result is not None:
            payload["result"] = result
        self._write_event(payload)

    def _start_cache_timer(self):
        self._cache_timer = QTimer(self)
        self._cache_timer.setInterval(self._cache_interval_ms)
        self._cache_timer.timeout.connect(self._sample_cache_progress)
        self._cache_timer.start()

    @staticmethod
    def _metric_call(obj, name):
        attr = getattr(obj, name, None)
        if attr is None:
            return None
        try:
            return attr() if callable(attr) else attr
        except Exception:
            return None

    def _collect_cache_state(self):
        state = {
            "current_frame": int(getattr(self.window, "current_frame", 0) or 0),
        }
        cache_obj = getattr(self.window, "cache_object", None)
        if cache_obj is not None:
            state["cache_class"] = cache_obj.__class__.__name__
            metrics = {}
            for name in (
                "Count",
                "GetCount",
                "Size",
                "GetSize",
                "Bytes",
                "GetBytes",
                "FrameCount",
                "GetFrameCount",
                "MaxBytes",
                "GetMaxBytes",
            ):
                value = self._metric_call(cache_obj, name)
                if value is not None:
                    metrics[name] = value
            if metrics:
                state["metrics"] = metrics

        preview_cache = Path(info.PREVIEW_CACHE_PATH)
        if preview_cache.exists():
            file_count = 0
            total_bytes = 0
            for child in preview_cache.iterdir():
                try:
                    if child.is_file():
                        file_count += 1
                        total_bytes += child.stat().st_size
                except Exception:
                    continue
            state["preview_cache_files"] = file_count
            state["preview_cache_bytes"] = total_bytes
        return state

    def _sample_cache_progress(self):
        if self._events_handle is None or self._closed:
            return
        try:
            state = self._collect_cache_state()
            signature = json.dumps(state, sort_keys=True, default=str)
            if signature == self._last_cache_signature:
                return
            self._last_cache_signature = signature
            payload = self._event_base("cache_progress")
            payload.update(state)
            self._write_event(payload)
        except Exception:
            log.debug("UI trace cache sampler failed", exc_info=1)

    def close(self):
        if self._closed:
            return
        self._closed = True

        if self._cache_timer:
            try:
                self._cache_timer.stop()
            except Exception:
                pass
            self._cache_timer = None

        try:
            if self._updates_listener_connected:
                get_app().updates.disconnect_listener(self)
        except Exception:
            pass
        self._updates_listener_connected = False

        try:
            if self._selection_connected:
                self.window.SelectionChanged.disconnect(self._on_selection_changed)
        except Exception:
            pass
        self._selection_connected = False

        if self._dialog_filter_installed:
            try:
                get_app().removeEventFilter(self)
            except Exception:
                pass
            self._dialog_filter_installed = False

        for action, slot in self._action_connections:
            try:
                action.triggered.disconnect(slot)
            except Exception:
                pass
        self._action_connections = []

        for signal, slot in self._dock_connections:
            try:
                signal.disconnect(slot)
            except Exception:
                pass
        self._dock_connections = []

        for slots in self._dialog_connections.values():
            for signal, slot in slots:
                try:
                    signal.disconnect(slot)
                except Exception:
                    pass
        self._dialog_connections.clear()

        for handle in (self._updates_handle, self._selections_handle, self._events_handle):
            try:
                if handle:
                    handle.flush()
                    handle.close()
            except Exception:
                pass
        self._updates_handle = None
        self._selections_handle = None
        self._events_handle = None



"""
 @file
 @brief This file contains the emojis listview, used by the main window
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from PyQt5.QtCore import QMimeData, QSize, QPoint, Qt, pyqtSlot, QRegExp
from PyQt5.QtGui import QDrag
from PyQt5.QtWidgets import QListView

import openshot  # Python module for libopenshot (required video editing module installed separately)
from classes import info
from classes.query import File
from classes.app import get_app
from classes.logger import log
import json
import uuid


class EmojisListView(QListView):
    """ A QListView QWidget used on the main window """
    drag_item_size = QSize(48, 48)
    drag_item_center = QPoint(24, 24)
    emoji_icon_size = QSize(75, 75)
    emoji_grid_size = QSize(80, 95)

    def dragEnterEvent(self, event):
        # If dragging urls onto widget, accept
        if event.mimeData().hasUrls():
            event.setDropAction(Qt.CopyAction)
            event.accept()

    def startDrag(self, event):
        """ Override startDrag method to display custom icon """

        # Get image of selected item
        selected = self.selectedIndexes()

        # Start drag operation
        drag = QDrag(self)
        drag.setMimeData(self.model.mimeData(selected))
        icon = self.model.data(selected[0], Qt.DecorationRole)
        drag.setPixmap(icon.pixmap(self.drag_item_size))
        drag.setHotSpot(self.drag_item_center)

        # Create emoji file before drag starts
        data = json.loads(drag.mimeData().text())

        # Get the translated emoji name from the model item
        # Map through proxy_model -> group_model -> standard model
        group_index = self.model.mapToSource(selected[0])
        source_index = self.group_model.mapToSource(group_index)
        selected_item = self.emojis_model.model.itemFromIndex(source_index)
        emoji_name = selected_item.text() if selected_item else None

        # Start a transaction so File + Clip are grouped for undo
        tid = str(uuid.uuid4())
        get_app().updates.transaction_id = tid

        file = self.add_file(data[0], emoji_name)

        # Update mimedata for emoji
        data = QMimeData()
        data.setText(json.dumps([file.id]))
        data.setHtml("clip")
        drag.setMimeData(data)

        # Start drag
        drag.exec_()

        # End transaction
        get_app().updates.transaction_id = None

    def add_file(self, filepath, emoji_name=None):
        # Add file into project

        app = get_app()
        _ = app._tr

        # Check for this path in our existing project data
        # ["1F595-1F3FE",
        # "openshot-qt-git/src/emojis/color/svg/1F595-1F3FE.svg"]
        file = File.get(path=filepath)

        # If this file is already found, exit
        if file:
            return file

        # Load filepath in libopenshot clip object (which will try multiple readers to open it)
        clip = openshot.Clip(filepath)

        # Get the JSON for the clip's internal reader
        try:
            reader = clip.Reader()
            file_data = json.loads(reader.Json())

            # Determine media type
            file_data["media_type"] = "image"

            # Set friendly emoji name (translated)
            if emoji_name:
                file_data["name"] = emoji_name

            # Save new file to the project data
            file = File()
            file.data = file_data
            file.save()
            return file

        except Exception as ex:
            # Log exception
            log.warning("Failed to import file: {}".format(str(ex)))

    @pyqtSlot(int)
    def group_changed(self, index=-1):
        emoji_group_name = get_app().window.emojiFilterGroup.itemText(index)
        emoji_group_id = get_app().window.emojiFilterGroup.itemData(index)
        self.group_model.setFilterFixedString(emoji_group_id)
        self.group_model.setFilterKeyColumn(2)

        # Save current emoji filter to settings
        s = get_app().get_settings()
        setting_emoji_group_id = s.get('emoji_group_filter') or 'smileys-emotion'
        if setting_emoji_group_id != emoji_group_id:
            s.set('emoji_group_filter', emoji_group_id)

        self.refresh_view()

    @pyqtSlot(str)
    def filter_changed(self, filter_text=None):
        """Filter emoji with proxy class"""

        self.model.setFilterRegExp(QRegExp(filter_text, Qt.CaseInsensitive))
        self.model.setFilterKeyColumn(0)
        self.refresh_view()

    def refresh_view(self):
        # Sort by column 0
        self.model.sort(0)

    def __init__(self, model):
        # Invoke parent init
        QListView.__init__(self)

        # Get external references
        app = get_app()
        _ = app._tr
        self.win = app.window

        # Get Model data
        self.emojis_model = model
        self.group_model = self.emojis_model.group_model
        self.model = self.emojis_model.proxy_model

        # Keep track of mouse press start position to determine when to start drag
        self.setAcceptDrops(True)
        self.setDragEnabled(True)
        self.setDropIndicatorShown(True)

        # Setup header columns and layout
        self.setModel(self.model)
        self.setIconSize(self.emoji_icon_size)
        self.setGridSize(self.emoji_grid_size)
        self.setViewMode(QListView.IconMode)
        self.setResizeMode(QListView.Adjust)
        self.setUniformItemSizes(True)
        self.setWordWrap(False)

        # Initialize sort
        self.refresh_view()

        # Get default emoji filter group
        s = get_app().get_settings()
        default_group_id = s.get('emoji_group_filter') or 'smileys-emotion'

        # setup filter events
        self.win.emojisFilter.textChanged.connect(self.filter_changed)

        # Loop through emoji groups, and populate emoji filter drop-down
        self.win.emojiFilterGroup.clear()
        self.win.emojiFilterGroup.addItem(_("Show All"), "")
        dropdown_index = 0
        for index, emoji_group_tuple in enumerate(sorted(self.emojis_model.emoji_groups)):
            emoji_group_name, emoji_group_id = emoji_group_tuple
            self.win.emojiFilterGroup.addItem(emoji_group_name, emoji_group_id)
            if emoji_group_id == default_group_id:
                # Initialize emoji filter group to settings
                # Off by one, due to 'show all' choice above
                dropdown_index = index + 1

        self.win.emojiFilterGroup.currentIndexChanged.connect(self.group_changed)
        self.win.emojiFilterGroup.setCurrentIndex(dropdown_index)



"""
 @file
 @brief This file loads the File Properties dialog
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import json

from PyQt5.QtWidgets import (
    QDialog, QFileDialog, QDialogButtonBox, QPushButton,
    )

# Python module for libopenshot (required video editing module installed separately)
import openshot

from uuid import uuid4
from classes import info, time_parts, ui_util
from classes.app import get_app
from classes.image_types import get_media_type
from classes.logger import log
from classes.metrics import track_metric_screen
from classes.query import Clip

MAX_FPS_SPINBOX_VALUE = 2147483647


class FileProperties(QDialog):
    """ File Properties Dialog """

    # Path to ui file
    ui_path = os.path.join(info.PATH, 'windows', 'ui', 'file-properties.ui')

    def __init__(self, file):
        self.file = file

        # Create dialog class
        QDialog.__init__(self)

        # Load UI from designer
        ui_util.load_ui(self, self.ui_path)

        # Init UI
        ui_util.init_ui(self)

        # get translations
        app = get_app()
        _ = app._tr

        # Get settings
        self.s = app.get_settings()

        # Track metrics
        track_metric_screen("file-properties-screen")

        # Add buttons to interface
        self.update_button = QPushButton(_('Update'))
        self.buttonBox.addButton(self.update_button, QDialogButtonBox.AcceptRole)
        self.buttonBox.addButton(QPushButton(_('Cancel')), QDialogButtonBox.RejectRole)

        # Dynamically load tabs from settings data
        self.settings_data = self.s.get_all_settings()

        # Initialize Form
        self.channel_layout_choices = []
        self.initialize()
        self.txtFrameRateNum.valueChanged.connect(self.update_frame_rate_display)
        self.txtFrameRateDen.valueChanged.connect(self.update_frame_rate_display)
        self.txtFrameRateNum.valueChanged.connect(self.update_duration_display)
        self.txtFrameRateDen.valueChanged.connect(self.update_duration_display)

    def initialize(self):
        """Init all form elements / textboxes / etc..."""
        # get translations
        app = get_app()
        _ = app._tr

        # Get file properties
        filename = os.path.basename(self.file.data["path"])
        file_extension = os.path.splitext(filename)[1]

        tags = self.file.data.get("tags", "")
        name = self.file.data.get("name", filename)

        # Populate fields
        self.txtFileName.setText(name)
        self.txtTags.setText(tags)
        self.txtFilePath.setText(self.file.data["path"])
        self.btnBrowse.clicked.connect(self.browsePath)

        # Populate video fields
        self.txtFrameRateNum.setMaximum(MAX_FPS_SPINBOX_VALUE)
        self.txtFrameRateDen.setMaximum(MAX_FPS_SPINBOX_VALUE)
        self.txtWidth.setValue(self.file.data["width"])
        self.txtHeight.setValue(self.file.data["height"])
        self.txtFrameRateNum.setValue(self.file.data["fps"]["num"])
        self.txtFrameRateDen.setValue(self.file.data["fps"]["den"])
        self.update_frame_rate_display()
        self.update_duration_display()
        self.txtAspectRatioNum.setValue(self.file.data["display_ratio"]["num"])
        self.txtAspectRatioDen.setValue(self.file.data["display_ratio"]["den"])
        self.txtPixelRatioNum.setValue(self.file.data["pixel_ratio"]["num"])
        self.txtPixelRatioDen.setValue(self.file.data["pixel_ratio"]["den"])

        # Only allow FPS edits for image-sequence style paths.
        fps_editable = "%" in str(self.file.data.get("path") or "")
        self.txtFrameRateNum.setEnabled(fps_editable)
        self.txtFrameRateDen.setEnabled(fps_editable)

        # Initialize start/end textboxes
        self.init_start_end_textboxes(self.file.data)

        # Populate video & audio format
        self.txtVideoFormat.setText(file_extension.replace(".", ""))
        self.txtVideoCodec.setText(self.file.data["vcodec"])
        self.txtAudioCodec.setText(self.file.data["acodec"])
        self.txtSampleRate.setValue(int(self.file.data["sample_rate"]))
        self.txtChannels.setValue(int(self.file.data["channels"]))
        self.txtVideoBitRate.setValue(int(self.file.data["video_bit_rate"]))
        self.txtAudioBitRate.setValue(int(self.file.data["audio_bit_rate"]))

        # Populate output field
        self.txtOutput.setText(json.dumps(self.file.data, sort_keys=True, indent=2))

        # Add channel layouts
        selected_channel_layout_index = 0
        current_channel_layout = 0
        if self.file.data["has_audio"]:
            current_channel_layout = int(self.file.data["channel_layout"])
        self.channel_layout_choices = []
        layouts = [(0, _("Unknown")),
                   (openshot.LAYOUT_MONO, _("Mono (1 Channel)")),
                   (openshot.LAYOUT_STEREO, _("Stereo (2 Channel)")),
                   (openshot.LAYOUT_SURROUND, _("Surround (3 Channel)")),
                   (openshot.LAYOUT_5POINT1, _("Surround (5.1 Channel)")),
                   (openshot.LAYOUT_7POINT1, _("Surround (7.1 Channel)"))]
        for channel_layout_index, layout in enumerate(layouts):
            log.info(layout)
            self.channel_layout_choices.append(layout[0])
            self.cboChannelLayout.addItem(layout[1], layout[0])
            if current_channel_layout == layout[0]:
                selected_channel_layout_index = channel_layout_index

        # Select matching channel layout
        self.cboChannelLayout.setCurrentIndex(selected_channel_layout_index)

        # Load the interlaced options
        self.cboInterlaced.clear()
        self.cboInterlaced.addItem(_("Yes"), "Yes")
        self.cboInterlaced.addItem(_("No"), "No")
        if self.file.data["interlaced_frame"]:
            self.cboInterlaced.setCurrentIndex(0)
        else:
            self.cboInterlaced.setCurrentIndex(1)

        # Switch to 1st page
        self.toolBox.setCurrentIndex(0)

    def update_frame_rate_display(self):
        """Show the current FPS fraction as a calculated float."""
        fps_den = self.txtFrameRateDen.value() or 1
        fps_float = self.txtFrameRateNum.value() / fps_den
        self.lblFrameRateValueDisplay.setText(f"= {fps_float:.2f}")

    def update_duration_display(self):
        """Show the duration adjusted to the currently entered FPS."""
        current_fps_den = self.txtFrameRateDen.value() or 1
        current_fps = self.txtFrameRateNum.value() / current_fps_den

        original_fps_meta = self.file.data.get("fps", {})
        original_fps_num = float(original_fps_meta.get("num") or 0.0)
        original_fps_den = float(original_fps_meta.get("den") or 1.0)
        original_fps = (
            original_fps_num / original_fps_den
            if original_fps_num > 0.0 and original_fps_den > 0.0
            else 0.0
        )

        duration_seconds = float(self.file.data.get("duration") or 0.0)
        if current_fps > 0.0 and original_fps > 0.0:
            duration_seconds *= original_fps / current_fps

        duration_parts = time_parts.secondsToTime(duration_seconds)
        self.txtDuration.setText(
            f"{duration_parts['hour']}:{duration_parts['min']}:{duration_parts['sec']}.{duration_parts['milli']}"
        )

    def init_start_end_textboxes(self, file_object):
        """Initialize the start and end textboxes based on a file object"""
        fps_float = float(file_object["fps"]["num"]) / float(file_object["fps"]["den"])

        self.txtStartFrame.setMaximum(int(file_object["video_length"]))
        if 'start' not in file_object.keys():
            self.txtStartFrame.setValue(1)
        else:
            self.txtStartFrame.setValue(round(float(file_object["start"]) * fps_float) + 1)

        self.txtEndFrame.setMaximum(int(file_object["video_length"]))
        if 'end' not in file_object.keys():
            self.txtEndFrame.setValue(int(file_object["video_length"]))
        else:
            # End times are stored as the first frame *after* the clip ends,
            # so convert to an inclusive frame number without adding 1.
            self.txtEndFrame.setValue(round(float(file_object["end"]) * fps_float))

    def verifyPath(self, new_path):
        """If the path has changed, verify that path is valid, and
        update duration, video_length, media_type, etc..."""

        # If this path could be an image sequence, get that info and prompt user.
        seq_info = get_app().window.files_model.get_image_sequence_details(new_path)
        get_app().window.files_model.ignore_image_sequence_paths = []

        # create the proper path for an image sequence
        if seq_info:
            # Override new_path with image sequence glob pattern
            new_path = seq_info.get("path")
            self.file.data["media_type"] = "video"

        # Open image sequence with Clip object (to determine metadata)
        clip = openshot.Clip(new_path)
        if clip and clip.info.duration > 0.0:
            # Make sure a clip can be created, then change the video length and path
            self.txtFilePath.setText(new_path)
            self.txtFileName.setText(os.path.basename(new_path))
            self.file.data = json.loads(clip.Reader().Json())
            if not seq_info:
                self.file.data["media_type"] = get_media_type(self.file.data)

            # Initialize start/end textboxes
            self.init_start_end_textboxes(self.file.data)
        else:
            log.info(f"Given path '{new_path}' was not a valid path... ignoring")

    def browsePath(self):
        # get translations
        app = get_app()
        _ = app._tr

        starting_folder, filename = os.path.split(self.file.data["path"])
        new_path = QFileDialog.getOpenFileName(None, _("Locate media file: %s") % filename, starting_folder)[0]

        # don't update if dialog was canceled
        if new_path:
            # verify path is valid (and prompt for image sequence if detected)
            self.verifyPath(new_path)

            # re-init form
            self.initialize()

    def accept(self):
        new_path = self.txtFilePath.text()
        if new_path and self.file.data.get("path") != new_path:
            # If path changed, verify path is valid (and prompt for image sequence if detected)
            self.verifyPath(new_path)

        # Update file details
        self.file.data["name"] = self.txtFileName.text()
        self.file.data["tags"] = self.txtTags.text()
        
        # Determine if FPS changed
        fps_float = self.txtFrameRateNum.value() / self.txtFrameRateDen.value()
        if self.file.data["fps"]["num"] != self.txtFrameRateNum.value() or \
                self.file.data["fps"]["den"] != self.txtFrameRateDen.value():
            original_fps_float = float(self.file.data["fps"]["num"]) / float(self.file.data["fps"]["den"])
            # Update file 'fps' and 'video_timebase'
            self.file.data["fps"]["num"] = self.txtFrameRateNum.value()
            self.file.data["fps"]["den"] = self.txtFrameRateDen.value()
            self.file.data["video_timebase"]["num"] = self.txtFrameRateDen.value()
            self.file.data["video_timebase"]["den"] = self.txtFrameRateNum.value()

            # Scale 'start' and 'end' properties by FPS difference
            fps_diff = original_fps_float / fps_float
            self.file.data["duration"] *= fps_diff
            if "start" in self.file.data:
                self.file.data["start"] *= fps_diff
            if "end" in self.file.data:
                self.file.data["end"] *= fps_diff

        # Scale 'start' and 'end' file attributes (if changed)
        elif self.txtStartFrame.value() != 1 or self.txtEndFrame.value() != int(self.file.data["video_length"]):
            # Scale 'start' and 'end' properties by FPS difference
            self.file.data["start"] = (self.txtStartFrame.value() - 1) / fps_float
            # End frames are inclusive, so convert to the time *after* the last frame
            self.file.data["end"] = self.txtEndFrame.value() / fps_float

        # Transaction id to group all updates together
        tid = str(uuid4())
        get_app().updates.transaction_id = tid

        # Save file object
        self.file.save()

        # Update file info & thumbnail
        get_app().window.FileUpdated.emit(self.file.id)

        # Update related clips
        for clip in Clip.filter(file_id=self.file.id):
            clip.data["reader"] = self.file.data
            clip.data["duration"] = self.file.data["duration"]
            if clip.data["end"] > clip.data["duration"]:
                clip.data["end"] = clip.data["duration"]
            clip.save()

            # Emit thumbnail update signal (to update timeline thumb image)
            thumbnail_frame = (clip.data["start"] * fps_float) + 1
            get_app().window.ThumbnailUpdated.emit(clip.id, thumbnail_frame)

        # Done grouping transactions
        get_app().updates.transaction_id = None

        # Accept dialog
        super(FileProperties, self).accept()

    def reject(self):

        # Cancel dialog
        super(FileProperties, self).reject()



"""
 @file
 @brief This file loads the UI for selecting a region of a video (rectangle used for effect processing)
  @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import sys
import functools
import math

from PyQt5.QtCore import *
from PyQt5.QtGui import QIcon, QPainter, QColor, QPen, QBrush, QKeySequence
from PyQt5.QtWidgets import *
import openshot  # Python module for libopenshot (required video editing module installed separately)

from classes import info, ui_util, time_parts, qt_types, updates
from classes.app import get_app
from classes.clip_utils import is_single_image_media
from classes.logger import log
from classes.metrics import *
from classes.proxy_service import dialog_preview_reader_data
from windows.preview_thread import PreviewParent
from windows.video_widget import VideoWidget

import json


class RegionAnnotatedSlider(QSlider):
    frameClicked = pyqtSignal(int)

    def __init__(self, orientation, parent=None):
        super().__init__(orientation, parent)
        self.setMouseTracking(True)
        self._total_frames = 1
        self._current_frame = 1
        self._markers = []  # list of (frame, kind)
        self._marker_positions = []  # list of (x, y, frame)

    def set_frames(self, total_frames, current_frame, markers):
        self._total_frames = int(max(1, total_frames or 1))
        self._current_frame = int(max(1, current_frame or 1))
        self._markers = sorted(list(markers or []), key=lambda item: int(item[0]))
        self.update()

    def _groove_rect(self):
        opt = QStyleOptionSlider()
        self.initStyleOption(opt)
        return self.style().subControlRect(QStyle.CC_Slider, opt, QStyle.SC_SliderGroove, self)

    def _handle_rect(self):
        opt = QStyleOptionSlider()
        self.initStyleOption(opt)
        return self.style().subControlRect(QStyle.CC_Slider, opt, QStyle.SC_SliderHandle, self)

    def _x_for_frame(self, frame):
        groove = self._groove_rect()
        left = float(groove.left())
        right = float(groove.right())
        span = max(1.0, right - left)
        if self._total_frames <= 1:
            return int(round(left))
        ratio = float(max(1, min(self._total_frames, int(frame))) - 1) / float(max(1, self._total_frames - 1))
        return int(round(left + ratio * span))

    def mousePressEvent(self, event):
        if event.button() != Qt.LeftButton:
            return super().mousePressEvent(event)
        handle = self._handle_rect().adjusted(-4, -4, 4, 4)
        if handle.contains(event.pos()):
            return super().mousePressEvent(event)
        if not self._marker_positions:
            return super().mousePressEvent(event)
        click_x = int(event.pos().x())
        click_y = int(event.pos().y())
        nearest = min(
            self._marker_positions,
            key=lambda item: (int(item[0]) - click_x) * (int(item[0]) - click_x) + (int(item[1]) - click_y) * (int(item[1]) - click_y),
        )
        dx = abs(int(nearest[0]) - click_x)
        dy = abs(int(nearest[1]) - click_y)
        if dx <= 8 and dy <= 8:
            self.setValue(int(nearest[2]))
            self.frameClicked.emit(int(nearest[2]))
            event.accept()
            return
        return super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        handle = self._handle_rect().adjusted(-4, -4, 4, 4)
        if handle.contains(event.pos()):
            self.unsetCursor()
            return super().mouseMoveEvent(event)
        if self._marker_positions:
            x = int(event.pos().x())
            y = int(event.pos().y())
            nearest = min(
                self._marker_positions,
                key=lambda item: (int(item[0]) - x) * (int(item[0]) - x) + (int(item[1]) - y) * (int(item[1]) - y),
            )
            if abs(int(nearest[0]) - x) <= 8 and abs(int(nearest[1]) - y) <= 8:
                self.setCursor(Qt.PointingHandCursor)
            else:
                self.unsetCursor()
        else:
            self.unsetCursor()
        return super().mouseMoveEvent(event)

    def leaveEvent(self, event):
        self.unsetCursor()
        return super().leaveEvent(event)

    def paintEvent(self, event):
        super().paintEvent(event)
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing, True)

        groove = self._groove_rect()
        # Center markers on the slider groove line.
        mid_y = int(groove.center().y())

        self._marker_positions = []
        for frame, kind in self._markers:
            x = self._x_for_frame(frame)
            self._marker_positions.append((x, mid_y, int(frame)))
            if kind == "both":
                painter.setPen(Qt.NoPen)
                painter.setBrush(QBrush(QColor("#53A0ED")))
                painter.drawEllipse(x - 4, mid_y - 4, 8, 8)
                painter.setBrush(QBrush(QColor("#E05757")))
                painter.drawEllipse(x, mid_y, 8, 8)
            elif kind == "negative":
                painter.setPen(Qt.NoPen)
                painter.setBrush(QBrush(QColor("#E05757")))
                painter.drawEllipse(x - 4, mid_y - 4, 8, 8)
            else:
                painter.setPen(Qt.NoPen)
                painter.setBrush(QBrush(QColor("#53A0ED")))
                painter.drawEllipse(x - 4, mid_y - 4, 8, 8)

        # Current-frame indicator
        cx = self._x_for_frame(self._current_frame)
        cur_pen = QPen(QColor("#EAF5FF"), 1)
        painter.setPen(cur_pen)
        painter.drawLine(cx, max(0, groove.top() - 5), cx, min(self.height() - 1, groove.bottom() + 5))
        painter.end()

class SelectRegion(QDialog):
    """ SelectRegion Dialog """

    # Path to ui file
    ui_path = os.path.join(info.PATH, 'windows', 'ui', 'region.ui')

    # Signals for preview thread
    previewFrameSignal = pyqtSignal(int)
    refreshFrameSignal = pyqtSignal()
    LoadFileSignal = pyqtSignal(str)
    PlaySignal = pyqtSignal()
    PauseSignal = pyqtSignal()
    SeekSignal = pyqtSignal(int)
    LoadTimelineAndSeekSignal = pyqtSignal(int)
    SpeedSignal = pyqtSignal(float)
    StopSignal = pyqtSignal()

    def __init__(self, file=None, clip=None, selection_mode="rect", parent=None):
        _ = get_app()._tr

        # Create dialog class
        QDialog.__init__(self, parent)

        # Load UI from designer
        ui_util.load_ui(self, self.ui_path)

        # Init UI
        ui_util.init_ui(self)
        self._esc_shortcut = QShortcut(QKeySequence("Esc"), self)
        self._esc_shortcut.setContext(Qt.WidgetWithChildrenShortcut)
        self._esc_shortcut.activated.connect(self.reject)
        if parent is None:
            self.setWindowFlags(
                (self.windowFlags() & ~Qt.Dialog)
                | Qt.Window
                | Qt.WindowMinMaxButtonsHint
                | Qt.WindowMaximizeButtonHint
            )
        else:
            self.setWindowModality(Qt.WindowModal)
        self.setSizeGripEnabled(True)

        # Track metrics
        track_metric_screen("cutting-screen")

        self.selection_mode = str(selection_mode or "rect").strip().lower()
        if self.selection_mode not in ("rect", "point", "annotate"):
            self.selection_mode = "rect"
        if self.selection_mode == "annotate":
            # Replace stock UI slider with custom-painted annotation slider
            # so marker dots are perfectly aligned to the slider groove.
            original_slider = self.sliderVideo
            custom_slider = RegionAnnotatedSlider(Qt.Horizontal, original_slider.parent())
            custom_slider.setObjectName(original_slider.objectName())
            custom_slider.setTracking(original_slider.hasTracking())
            if hasattr(self, "horizontalLayout_3"):
                self.horizontalLayout_3.replaceWidget(original_slider, custom_slider)
            original_slider.hide()
            original_slider.deleteLater()
            self.sliderVideo = custom_slider
            self.sliderVideo.frameClicked.connect(self._on_marker_frame_clicked)
        self._selected_points = []
        self._selected_points_negative = []
        self._selected_payload = {}
        self._selected_rect_normalized = None
        self._selected_region_qimage = None
        self.loop_playback = False
        self.frame_annotations = {}
        self._last_annotation_frame = 1
        self._frame_has_local_keyframe = False
        self._frame_edited = False

        self.start_frame = 1
        self.start_image = None
        self.end_frame = 1
        self.end_image = None
        self.current_frame = 1
        self.file = file
        self.reader_data = dialog_preview_reader_data(file) if file else {}
        self.file_path = str(self.reader_data.get("path") or "")
        is_single_image = bool(is_single_image_media(self.reader_data) or is_single_image_media(getattr(file, "data", {})))

        # Create region clip with Reader
        if clip:
            if self.file_path:
                self.clip = openshot.Clip(self.file_path)
                if not is_single_image:
                    self.clip.SetJson(json.dumps({"reader": self.reader_data}))
            else:
                self.clip = openshot.Clip(clip.Reader())
            self.clip.Open()
            # Set region clip start and end
            self.clip.Start(clip.Start())
            self.clip.End(clip.End())
        else:
            source_path = self.file_path
            if not source_path and file:
                if hasattr(file, "absolute_path"):
                    source_path = file.absolute_path()
                else:
                    source_path = str(getattr(file, "data", {}).get("path", ""))
            self.clip = openshot.Clip(source_path)
            if self.reader_data and not is_single_image:
                self.clip.SetJson(json.dumps({"reader": self.reader_data}))
            self.clip.Open()
        self.clip.Id(get_app().project.generate_id())

        if not self.file_path:
            if file and hasattr(file, "absolute_path"):
                self.file_path = file.absolute_path()
            else:
                self.file_path = str(getattr(file, "data", {}).get("path", ""))

        c_info = self.clip.Reader().info
        self.fps = c_info.fps.ToInt()
        self.fps_num = c_info.fps.num
        self.fps_den = c_info.fps.den
        self.width = c_info.width
        self.height = c_info.height
        self.sample_rate = int(c_info.sample_rate)
        self.channels = int(c_info.channels)
        self.channel_layout = int(c_info.channel_layout)
        self.video_length = int(self.clip.Duration() * self.fps) + 1

        # Apply effects to region frames
        if clip:
            for effect in clip.Effects():
                self.clip.AddEffect(effect)

        # Open video file with Reader
        log.info(self.clip.Reader())

        # Set instruction text first so it remains above the preview widget.
        if self.selection_mode == "point":
            self.lblInstructions.setText(
                _("Click to add tracking point (SHIFT+Click for additional points, CTRL+Click for negative point)")
            )
        elif self.selection_mode == "annotate":
            self.lblInstructions.setText(
                _("Choose a tool and mark positive/negative points or rectangles. Scrub to edit selections by frame.")
            )
        else:
            self.lblInstructions.setText(_("Draw a rectangle to select a region of the video frame."))

        # Add Video Widget
        self.videoPreview = VideoWidget(watch_project=False)
        self.videoPreview.win = self
        self.videoPreview.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        self.videoPreview.region_selection_mode = self.selection_mode
        self.videoPreview.regionAnnotationChanged.connect(self._on_video_annotation_changed)
        self.verticalLayout.insertWidget(1, self.videoPreview)
        if self.selection_mode == "annotate":
            self._build_annotation_toolbar()

        # Set aspect ratio to match source content
        aspect_ratio = openshot.Fraction(self.width, self.height)
        aspect_ratio.Reduce()
        self.videoPreview.aspect_ratio = aspect_ratio

        # Set max size of video preview (for speed)
        self.viewport_rect = self.videoPreview.centeredViewport(self.width, self.height)

        # Create an instance of a libopenshot Timeline object
        self.r = openshot.Timeline(max(2, int(self.width or self.viewport_rect.width() or 2)),
                                   max(2, int(self.height or self.viewport_rect.height() or 2)),
                                   openshot.Fraction(self.fps_num, self.fps_den),
                                   self.sample_rate, self.channels, self.channel_layout)
        self.r.info.channel_layout = self.channel_layout
        self.r.SetMaxSize(self.viewport_rect.width(), self.viewport_rect.height())

        try:
            # Show waveform for audio files
            if not self.clip.Reader().info.has_video and self.clip.Reader().info.has_audio:
                self.clip.Waveform(True)

            # Set has_audio property
            self.r.info.has_audio = self.clip.Reader().info.has_audio

            # Update video_length property of the Timeline object
            self.r.info.video_length = self.video_length

            self.r.AddClip(self.clip)

        except:
            log.error('Failed to load media file into region select player: %s' % self.file_path)
            return

        # Open reader
        self.r.Open()

        # Start the preview thread
        self.initialized = False
        self.transforming_clip = False
        self.preview_parent = PreviewParent()
        self.preview_parent.Init(self, self.r, self.videoPreview, self.video_length)
        self.preview_thread = self.preview_parent.worker

        # Set slider constraints
        self.sliderIgnoreSignal = False
        self.sliderVideo.setMinimum(1)
        self.sliderVideo.setMaximum(self.video_length)
        self.sliderVideo.setSingleStep(1)
        self.sliderVideo.setPageStep(24)
        self.videoPreview.delayed_resize_timer.timeout.connect(self._apply_dynamic_preview_max_size)

        # Display start frame (and then the previous frame)
        QTimer.singleShot(500, functools.partial(self.sliderVideo.setValue, 2))
        QTimer.singleShot(600, functools.partial(self.sliderVideo.setValue, 1))

        # Add buttons
        self.cancel_button = QPushButton(_('Cancel'))
        if self.selection_mode == "rect":
            process_label = _('Select Region')
        elif self.selection_mode == "annotate":
            process_label = _('Apply Selections')
        else:
            process_label = _('Select Point(s)')
        self.process_button = QPushButton(process_label)
        self.buttonBox.addButton(self.process_button, QDialogButtonBox.AcceptRole)
        self.buttonBox.addButton(self.cancel_button, QDialogButtonBox.RejectRole)

        # Connect signals
        self.actionPlay.triggered.connect(self.actionPlay_Triggered)
        self.btnPlay.clicked.connect(self.btnPlay_clicked)
        self.sliderVideo.valueChanged.connect(self.sliderVideo_valueChanged)
        self.initialized = True
        if self.selection_mode == "annotate":
            self._load_frame_annotations(1)
            self._update_defined_frames_label()
            self._refresh_marker_bar()

        get_app().window.SelectRegionSignal.emit(self.clip.Id())

    def actionPlay_Triggered(self):
        # Trigger play button (This action is invoked from the preview thread, so it must exist here)
        self.btnPlay.click()

    def keyPressEvent(self, event):
        if event and event.key() == Qt.Key_Escape:
            self.reject()
            event.accept()
            return
        return super(SelectRegion, self).keyPressEvent(event)

    def _target_preview_max_size(self):
        viewport_rect = self.videoPreview.centeredViewport(self.videoPreview.width(), self.videoPreview.height())
        device_pixel_ratio = self.devicePixelRatioF()
        requested = QSize(
            max(2, int(round(viewport_rect.width() * device_pixel_ratio))),
            max(2, int(round(viewport_rect.height() * device_pixel_ratio))),
        )

        source_width = int(getattr(self, "width", 0) or 0)
        source_height = int(getattr(self, "height", 0) or 0)
        file_data = getattr(getattr(self, "file", None), "data", {})
        if is_single_image_media(getattr(self, "reader_data", {})) or is_single_image_media(file_data):
            project = getattr(get_app(), "project", None)
            if project:
                source_width = max(source_width, int(project.get("width") or 0))
                source_height = max(source_height, int(project.get("height") or 0))
        if source_width > 0 and source_height > 0:
            source_size = QSize(source_width, source_height)
            if requested.width() > source_width or requested.height() > source_height:
                capped = QSize(source_size)
            else:
                capped = QSize(requested)
        else:
            capped = requested

        if capped.height() > 0:
            ratio = float(capped.width()) / float(capped.height())
            even_width = max(2, int(round(capped.width() / 2.0) * 2))
            even_height = max(2, int(round(round(even_width / ratio) / 2.0) * 2))
            capped = QSize(even_width, even_height)

        return capped

    def _apply_dynamic_preview_max_size(self):
        if not getattr(self, "initialized", False) or not getattr(self, "r", None):
            return

        new_size = self._target_preview_max_size()
        previous_width = int(getattr(self.r, "preview_width", 0) or 0)
        previous_height = int(getattr(self.r, "preview_height", 0) or 0)

        if previous_width == new_size.width() and previous_height == new_size.height():
            return

        self.PauseSignal.emit()
        self.r.SetMaxSize(new_size.width(), new_size.height())
        self.r.ClearAllCache(True)
        self.viewport_rect = self.videoPreview.centeredViewport(self.width, self.height)
        self.refreshFrameSignal.emit()

    def _icon_path(self, name):
        icon_name = str(name or "").strip()
        path = os.path.join(info.PATH, "themes", "humanity", "images", icon_name)
        if os.path.exists(path):
            return path
        return ""

    def _build_annotation_toolbar(self):
        _ = get_app()._tr
        self.annotation_toolbar = QHBoxLayout()
        self.annotation_toolbar.setContentsMargins(0, 0, 0, 0)
        self.annotation_toolbar.setSpacing(6)

        self.annotation_tool_group = QButtonGroup(self)
        self.annotation_tool_group.setExclusive(True)
        checked_style = (
            "QToolButton:checked {"
            "  background-color: palette(highlight);"
            "  color: palette(highlighted-text);"
            "  border: 1px solid palette(highlight);"
            "}"
        )

        tool_defs = [
            ("positive_point", _("Positive Point"), "ai-track-point-positive.svg"),
            ("negative_point", _("Negative Point"), "ai-track-point-negative.svg"),
            ("positive_rect", _("Positive Rectangle"), "ai-track-rect-positive.svg"),
            ("negative_rect", _("Negative Rectangle"), "ai-track-rect-negative.svg"),
        ]
        self.annotation_tool_buttons = {}
        for tool_id, tooltip, icon_name in tool_defs:
            btn = QToolButton(self)
            btn.setCheckable(True)
            btn.setToolTip(tooltip)
            btn.setIconSize(QSize(18, 18))
            btn.setMinimumSize(QSize(28, 28))
            icon_path = self._icon_path(icon_name)
            if icon_path:
                btn.setIcon(QIcon(icon_path))
            else:
                btn.setText(tooltip)
            btn.setStyleSheet(checked_style)
            btn.clicked.connect(lambda checked=False, t=tool_id: self._on_annotation_tool_changed(t))
            self.annotation_tool_group.addButton(btn)
            self.annotation_tool_buttons[tool_id] = btn
            self.annotation_toolbar.addWidget(btn)

        self.btnClearAnnotation = QToolButton(self)
        self.btnClearAnnotation.setToolTip(_("Clear All Selections"))
        self.btnClearAnnotation.setIconSize(QSize(18, 18))
        self.btnClearAnnotation.setMinimumSize(QSize(28, 28))
        trash_icon = self._icon_path("track-delete-enabled.svg")
        if trash_icon:
            self.btnClearAnnotation.setIcon(QIcon(trash_icon))
        else:
            self.btnClearAnnotation.setText(_("Reset"))
        self.btnClearAnnotation.clicked.connect(self._clear_current_frame_annotations)
        self.annotation_toolbar.addSpacing(8)
        self.annotation_toolbar.addWidget(self.btnClearAnnotation)
        self.annotation_toolbar.addStretch(1)

        self.lblDefinedFrames = QLabel("")
        self.annotation_toolbar.addWidget(self.lblDefinedFrames)
        self.verticalLayout.insertLayout(1, self.annotation_toolbar)

        # Default tool
        default_btn = self.annotation_tool_buttons.get("positive_point")
        if default_btn:
            default_btn.setChecked(True)
        self._on_annotation_tool_changed("positive_point")

    def _on_annotation_tool_changed(self, tool_id):
        if hasattr(self, "videoPreview") and self.videoPreview is not None:
            self.videoPreview.region_annotation_tool = str(tool_id or "positive_point")

    def _capture_current_annotation(self):
        def _points_to_payload(items):
            payload = []
            for p in items or []:
                try:
                    payload.append({"x": float(p.x()), "y": float(p.y())})
                except Exception:
                    continue
            return payload

        def _rects_to_payload(items):
            payload = []
            for r in items or []:
                if not isinstance(r, QRectF):
                    continue
                n = r.normalized()
                payload.append({
                    "x1": float(n.left()),
                    "y1": float(n.top()),
                    "x2": float(n.right()),
                    "y2": float(n.bottom()),
                })
            return payload

        return {
            "positive_points": _points_to_payload(self.videoPreview.region_points_positive),
            "negative_points": _points_to_payload(self.videoPreview.region_points_negative),
            "positive_rects": _rects_to_payload(self.videoPreview.region_rects_positive),
            "negative_rects": _rects_to_payload(self.videoPreview.region_rects_negative),
        }

    def _has_any_annotation(self, payload):
        return bool(
            (payload.get("positive_points") or [])
            or (payload.get("negative_points") or [])
            or (payload.get("positive_rects") or [])
            or (payload.get("negative_rects") or [])
        )

    def _save_current_frame_annotations(self, force=False):
        if self.selection_mode != "annotate":
            return
        frame = int(max(1, self.current_frame))
        if (not force) and (not self._frame_has_local_keyframe) and (not self._frame_edited):
            return
        payload = self._capture_current_annotation()
        if self._has_any_annotation(payload):
            self.frame_annotations[frame] = payload
        elif frame in self.frame_annotations:
            self.frame_annotations.pop(frame, None)
        self._frame_has_local_keyframe = frame in self.frame_annotations
        self._frame_edited = False
        self._update_defined_frames_label()

    def _load_frame_annotations(self, frame):
        if self.selection_mode != "annotate":
            return
        frame = int(max(1, frame))
        payload = {}
        inherited = False
        if frame in self.frame_annotations:
            payload = dict(self.frame_annotations.get(frame, {}))
        else:
            prior_frames = [f for f in self.frame_annotations.keys() if int(f) <= frame]
            if prior_frames:
                nearest = int(sorted(prior_frames)[-1])
                payload = dict(self.frame_annotations.get(nearest, {}))
                inherited = True
        self.videoPreview.region_points_positive = [
            QPointF(float(p.get("x", 0.0)), float(p.get("y", 0.0)))
            for p in (payload.get("positive_points") or [])
            if isinstance(p, dict)
        ]
        self.videoPreview.region_points_negative = [
            QPointF(float(p.get("x", 0.0)), float(p.get("y", 0.0)))
            for p in (payload.get("negative_points") or [])
            if isinstance(p, dict)
        ]
        self.videoPreview.region_rects_positive = [
            QRectF(
                QPointF(float(r.get("x1", 0.0)), float(r.get("y1", 0.0))),
                QPointF(float(r.get("x2", 0.0)), float(r.get("y2", 0.0))),
            ).normalized()
            for r in (payload.get("positive_rects") or [])
            if isinstance(r, dict)
        ]
        self.videoPreview.region_rects_negative = [
            QRectF(
                QPointF(float(r.get("x1", 0.0)), float(r.get("y1", 0.0))),
                QPointF(float(r.get("x2", 0.0)), float(r.get("y2", 0.0))),
            ).normalized()
            for r in (payload.get("negative_rects") or [])
            if isinstance(r, dict)
        ]
        self.videoPreview.region_rect_drag_start = None
        self.videoPreview.region_rect_drag_current = None
        self.videoPreview.region_annotation_inherited = bool(inherited)
        self.videoPreview.update()
        self._frame_has_local_keyframe = frame in self.frame_annotations
        self._frame_edited = False

    def _clear_current_frame_annotations(self):
        if self.selection_mode != "annotate":
            return
        self.frame_annotations = {}
        self._frame_has_local_keyframe = False
        self._frame_edited = False
        self.videoPreview.region_points_positive = []
        self.videoPreview.region_points_negative = []
        self.videoPreview.region_rects_positive = []
        self.videoPreview.region_rects_negative = []
        self.videoPreview.region_rect_drag_start = None
        self.videoPreview.region_rect_drag_current = None
        self.videoPreview.region_annotation_inherited = False
        self.videoPreview.update()
        self._update_defined_frames_label()
        self._refresh_marker_bar()

    def _on_video_annotation_changed(self):
        if self.selection_mode != "annotate":
            return
        self._frame_edited = True
        self._save_current_frame_annotations(force=True)
        self._refresh_marker_bar()

    def _update_defined_frames_label(self):
        if self.selection_mode != "annotate" or not hasattr(self, "lblDefinedFrames"):
            return
        frames = sorted(self.frame_annotations.keys())
        if not frames:
            self.lblDefinedFrames.setText("")
            return
        preview = ", ".join(str(f) for f in frames[:10])
        if len(frames) > 10:
            preview = "{} ...".format(preview)
        self.lblDefinedFrames.setText(get_app()._tr("Frames: {}").format(preview))
        self._refresh_marker_bar()

    def _refresh_marker_bar(self):
        if self.selection_mode != "annotate" or not hasattr(self, "sliderVideo"):
            return
        if not hasattr(self.sliderVideo, "set_frames"):
            return
        markers = []
        for frame in sorted(self.frame_annotations.keys()):
            payload = self.frame_annotations.get(frame, {}) or {}
            has_pos = bool((payload.get("positive_points") or []) or (payload.get("positive_rects") or []))
            has_neg = bool((payload.get("negative_points") or []) or (payload.get("negative_rects") or []))
            kind = "both" if (has_pos and has_neg) else ("negative" if has_neg else "positive")
            markers.append((int(frame), kind))
        self.sliderVideo.set_frames(self.video_length, self.current_frame, markers)

    def _on_marker_frame_clicked(self, frame_number):
        frame_number = int(max(1, min(int(frame_number), int(self.video_length))))
        self.sliderVideo.setValue(frame_number)

    def selection_payload(self):
        return dict(self._selected_payload or {})

    def selected_rect_normalized(self):
        if isinstance(self._selected_rect_normalized, dict):
            return dict(self._selected_rect_normalized)
        return None

    def selected_region_qimage(self):
        return self._selected_region_qimage

    def movePlayhead(self, frame_number):
        """Update the playhead position"""

        if self.selection_mode == "annotate" and int(frame_number) != int(self.current_frame):
            self._save_current_frame_annotations()
        self.current_frame = frame_number
        # Move slider to correct frame position
        self.sliderIgnoreSignal = True
        self.sliderVideo.setValue(frame_number)
        self.sliderIgnoreSignal = False

        # Convert frame to seconds
        seconds = (frame_number-1) / self.fps

        # Convert seconds to time stamp
        time_text = time_parts.secondsToTime(seconds, self.fps_num, self.fps_den)
        timestamp = "%s:%s:%s:%s" % (time_text["hour"], time_text["min"], time_text["sec"], time_text["frame"])

        # Update label
        self.lblVideoTime.setText(timestamp)
        if self.selection_mode == "annotate":
            self._load_frame_annotations(frame_number)
            self._refresh_marker_bar()

    def btnPlay_clicked(self, force=None):
        log.info("btnPlay_clicked")

        if force == "pause":
            self.btnPlay.setChecked(False)
        elif force == "play":
            self.btnPlay.setChecked(True)

        if self.btnPlay.isChecked():
            log.info('play (icon to pause)')
            ui_util.setup_icon(self, self.btnPlay, "actionPlay", "media-playback-pause")
            self.preview_thread.Play()
        else:
            log.info('pause (icon to play)')
            ui_util.setup_icon(self, self.btnPlay, "actionPlay", "media-playback-start")  # to default
            self.preview_thread.Pause()

        # Send focus back to toolbar
        self.sliderVideo.setFocus()

    def sliderVideo_valueChanged(self, new_frame):
        if self.preview_thread and not self.sliderIgnoreSignal:
            log.info('sliderVideo_valueChanged: %s' % new_frame)
            if self.selection_mode == "annotate":
                self._save_current_frame_annotations()
                self.current_frame = int(new_frame)
                self._load_frame_annotations(new_frame)
                self._refresh_marker_bar()

            # Pause video
            self.btnPlay_clicked(force="pause")

            # Seek to new frame
            self.preview_thread.previewFrame(new_frame)

    def accept(self):
        """ Ok button clicked """
        # get translations
        app = get_app()
        _ = app._tr

        # Legacy behavior for rect/point modes: require frame 1 selection.
        if self.selection_mode in ("rect", "point") and self.sliderVideo.value() != self.sliderVideo.minimum():
            # Show a warning message box to the user
            QMessageBox.warning(self, _("Invalid Region"),
                                _("Please choose a region at the beginning of the clip"))

            # Reset the slider to its minimum value
            self.sliderVideo.setValue(self.sliderVideo.minimum())
            return

        if self.selection_mode == "point" and not self.videoPreview.region_points_positive:
            QMessageBox.warning(self, _("Invalid Selection"), _("Please select at least one point."))
            return
        if self.selection_mode == "rect":
            top_left = getattr(self.videoPreview, "regionTopLeftHandle", None)
            bottom_right = getattr(self.videoPreview, "regionBottomRightHandle", None)
            if top_left is None or bottom_right is None:
                QMessageBox.warning(self, _("Invalid Selection"), _("Please draw a rectangle region."))
                return
            curr_frame_size = getattr(self.videoPreview, "curr_frame_size", None)
            if curr_frame_size and curr_frame_size.width() > 0 and curr_frame_size.height() > 0:
                x1 = float(top_left.x()) / float(curr_frame_size.width())
                y1 = float(top_left.y()) / float(curr_frame_size.height())
                x2 = float(bottom_right.x()) / float(curr_frame_size.width())
                y2 = float(bottom_right.y()) / float(curr_frame_size.height())
                left = min(x1, x2)
                top = min(y1, y2)
                right = max(x1, x2)
                bottom = max(y1, y2)
                self._selected_rect_normalized = {
                    "normalized_x": left,
                    "normalized_y": top,
                    "normalized_width": max(0.0, right - left),
                    "normalized_height": max(0.0, bottom - top),
                }
            region_qimage = getattr(self.videoPreview, "region_qimage", None)
            if region_qimage:
                self._selected_region_qimage = region_qimage.copy()
        if self.selection_mode == "annotate":
            self._save_current_frame_annotations()
            if not self.frame_annotations:
                QMessageBox.warning(self, _("Invalid Selection"), _("Please select at least one point or rectangle."))
                return
            sorted_frames = sorted(self.frame_annotations.keys())
            seed_frame = int(sorted_frames[0]) if sorted_frames else int(self.current_frame)
            self._selected_payload = {
                "version": 1,
                "seed_frame": seed_frame,
                "frames": {
                    str(frame): dict(self.frame_annotations.get(frame, {}))
                    for frame in sorted_frames
                },
            }
        else:
            self._selected_payload = {}

        # Continue with the rest of the accept method
        self._selected_points = self.selected_points()
        self._selected_points_negative = self.selected_points_negative()
        self.shutdownPlayer()
        get_app().window.SelectRegionSignal.emit("")
        super(SelectRegion, self).accept()

    def shutdownPlayer(self):

        log.info('shutdownPlayer')

        # Stop playback
        self.preview_parent.Stop()

        # Close readers
        self.clip.Close()
        # self.r.RemoveClip(self.clip)
        self.r.Close()
        # self.clip.Close()
        self.r.ClearAllCache()

    def reject(self):

        # Cancel dialog
        self.shutdownPlayer()
        get_app().window.SelectRegionSignal.emit("")
        super(SelectRegion, self).reject()

    def selected_points(self):
        if self._selected_points:
            return list(self._selected_points)
        points = []
        for point in getattr(self.videoPreview, "region_points_positive", []) or []:
            points.append({"x": float(point.x()), "y": float(point.y())})
        return points

    def selected_points_negative(self):
        if self._selected_points_negative:
            return list(self._selected_points_negative)
        points = []
        for point in getattr(self.videoPreview, "region_points_negative", []) or []:
            points.append({"x": float(point.x()), "y": float(point.y())})
        return points



"""
 @file
 @brief Finite state machine for timeline interactions
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from PyQt5.QtCore import QState, QStateMachine


class DragState(QState):
    def __init__(self, widget):
        super().__init__()
        self.widget = widget

    def onEntry(self, event):
        self.widget._startClipDrag()

    def onExit(self, event):
        self.widget._finishClipDrag()


class ResizeState(QState):
    def __init__(self, widget):
        super().__init__()
        self.widget = widget

    def onEntry(self, event):
        self.widget._startResize()

    def onExit(self, event):
        self.widget._finishResize()


class PlayheadState(QState):
    def __init__(self, widget):
        super().__init__()
        self.widget = widget

    def onEntry(self, event):
        self.widget._startPlayhead()

    def onExit(self, event):
        self.widget._finishPlayhead()


class BoxSelectState(QState):
    def __init__(self, widget):
        super().__init__()
        self.widget = widget

    def onEntry(self, event):
        self.widget._startBoxSelect()

    def onExit(self, event):
        self.widget._finishBoxSelect()


class KeyframeState(QState):
    def __init__(self, widget):
        super().__init__()
        self.widget = widget

    def onEntry(self, event):
        self.widget._startKeyframeDrag()

    def onExit(self, event):
        self.widget._finishKeyframeDrag()


class TimelineStateMachine(QStateMachine):
    def __init__(self, widget):
        super().__init__(widget)
        self.idle = QState()
        self.drag = DragState(widget)
        self.resize = ResizeState(widget)
        self.playhead = PlayheadState(widget)
        self.box = BoxSelectState(widget)
        self.keyframe = KeyframeState(widget)

        # States must be registered with the machine before they can be used as
        # the initial state.  Failing to do so causes Qt to warn that the
        # initial state is not a child of the machine, and on Windows it
        # eventually leads to a crash when the state machine starts.
        for state in (self.idle, self.drag, self.resize, self.playhead, self.box, self.keyframe):
            self.addState(state)

        self.setInitialState(self.idle)
        self.start()



"""
 @file
 @brief Painter for clip and transition keyframes.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

from PyQt5.QtCore import QRectF
from PyQt5.QtGui import QColor, QPainter, QPainterPath, QPen

from .base import BasePainter


class KeyframePainter(BasePainter):
    def update_theme(self):
        fill = self.w.theme.keyframe_fill
        border = self.w.theme.keyframe_border
        base_color = QColor("#4d7bff")
        if fill.isValid():
            base_color = QColor(fill)
        self.fill = base_color
        border_color = QColor("#ffffff")
        if border.isValid():
            border_color = QColor(border)
        self.border = border_color
        self.pen = QPen(self.border, 1.2)
        self.pen.setCosmetic(True)
        self.inactive_opacity = getattr(self.w.theme, "keyframe_inactive_opacity", 0.5)
        self.size = max(1, int(getattr(self.w.theme, "keyframe_size", 10) or 10))

    def paint(self, painter: QPainter):
        markers = getattr(self.w, "_keyframe_markers", [])
        if not markers:
            return

        painter.save()
        painter.setRenderHint(QPainter.Antialiasing, True)
        painter.setPen(self.pen)
        for marker in markers:
            rect = marker.get("rect")
            if not isinstance(rect, QRectF) or rect.isNull():
                continue
            opacity = 1.0 if marker.get("selected") else self.inactive_opacity
            if marker.get("dimmed"):
                opacity *= 0.5
            painter.setOpacity(opacity)
            color = marker.get("color")
            if not isinstance(color, QColor) or not color.isValid():
                color = self.fill
            painter.setBrush(color)
            interpolation = marker.get("interpolation", "bezier")
            if interpolation == "linear":
                painter.drawRect(rect)
            elif interpolation == "constant":
                path = QPainterPath()
                center = rect.center()
                path.moveTo(center.x(), rect.top())
                path.lineTo(rect.right(), center.y())
                path.lineTo(center.x(), rect.bottom())
                path.lineTo(rect.left(), center.y())
                path.closeSubpath()
                painter.drawPath(path)
            else:
                painter.drawEllipse(rect)
        painter.restore()



"""
 @file
 @brief Utility functions for manipulating SVG style attributes
 @author FeRD (Frank Dana) <ferdnyc@gmail.com>

 @section LICENSE

 Copyright (c) 2008-2020 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
"""

from classes.logger import log


def style_to_dict(style: str) -> dict:
    """Explode an SVG node style= attribute string into a dict representation"""
    styledict = {}
    try:
        # Fill the dict using key-value pairs produced by this comprehension
        styledict.update(
            # Make pairs of (property, value) by splitting on the first ':'
            (a.split(':', 1))
            # Using the list formed by splitting style on ';'
            for a in style.split(';')
            # Ignore any empty strings produced
            if a
            )
        return styledict
    except ValueError as ex:
        log.error(
            "style_to_dict failed to convert to dict: %s\n%s",
            ex, style)


def dict_to_style(styledict: dict) -> str:
    """Turn an exploded style dictionary back into a string"""
    # Glue the results of this comprehension together with ';' separators
    try:
        style = ";".join([
            # Produce a list of "key:value" strings
            ":".join([k, v])
            # For every {key: value} in styledict
            for k, v in styledict.items()
            ])
        # Don't forget the trailing semicolon!
        return style + ';'
    except ValueError as ex:
        import json
        log.error(
            "style_to_dict failed to generate string: %s\n%s",
            ex, json.dumps(styledict))


def set_if_existing(d: dict, existing_key, new_value):
    if existing_key in d:
        d.update({existing_key: new_value})



"""
 @file
 @brief This file contains the project file treeview, used by the main window
 @author Noah Figg <eggmunkee@hotmail.com>
 @author Jonathan Thomas <jonathan@openshot.org>
 @author Olivier Girard <eolinwen@gmail.com>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import uuid

from PyQt5.QtCore import QSize, Qt, QPoint, QItemSelectionModel
from PyQt5.QtGui import QDrag, QCursor, QPixmap, QPainter, QIcon, QColor, QFontMetrics
from PyQt5.QtWidgets import QTreeView, QAbstractItemView, QSizePolicy, QHeaderView, QStyledItemDelegate, QStyleOptionViewItem, QStyle

from classes import info
from classes.app import get_app
from classes.logger import log
from classes.query import File
from classes.qt_types import font_metrics_horizontal_advance
from .ai_tools_menu import add_ai_tools_menu
from .files_thumbnail_overlay import paint_media_overlay, paint_proxy_badge
from .menu import StyledContextMenu
from .optimized_preview_menu import add_optimized_preview_menu


def _is_generation_placeholder(file_id):
    return str(file_id or "").startswith("__genjob__:")


def _job_id_from_placeholder(file_id):
    file_id = str(file_id or "")
    if not _is_generation_placeholder(file_id):
        return None
    return file_id.split(":", 1)[1]


class FilesTreeProgressDelegate(QStyledItemDelegate):
    """Paint a thin progress line over thumbnail cells."""

    def __init__(self, view):
        super().__init__(view)
        self.view = view

    def paint(self, painter, option, index):
        super().paint(painter, option, index)

        if index.column() != 0:
            return

        opt = QStyleOptionViewItem(option)
        self.initStyleOption(opt, index)
        style = opt.widget.style() if opt.widget else self.view.style()
        deco_rect = style.subElementRect(QStyle.SE_ItemViewItemDecoration, opt, opt.widget)
        if not deco_rect.isValid():
            return

        media_type = index.sibling(index.row(), 3).data(Qt.DisplayRole)
        paint_media_overlay(painter, deco_rect, media_type)

        file_id = index.sibling(index.row(), 5).data(Qt.DisplayRole)
        queue = getattr(self.view.win, "generation_queue", None)
        generation_badge = None
        if file_id and queue:
            generation_badge = queue.get_file_badge(file_id)
            if not generation_badge and _is_generation_placeholder(file_id):
                job = queue.get_job(_job_id_from_placeholder(file_id))
                if job and job.get("status") in ("queued", "running", "canceling"):
                    label = "Queued" if job.get("status") == "queued" else "Generating"
                    generation_badge = {
                        "status": job.get("status"),
                        "progress": int(job.get("progress", 0)),
                        "label": label,
                        "job_id": job.get("id"),
                    }

        proxy_badge = None
        proxy_service = getattr(self.view.win, "proxy_service", None)
        if file_id and proxy_service and not _is_generation_placeholder(file_id):
            proxy_badge = proxy_service.get_file_badge(file_id)
            file_obj = File.get(id=file_id)
            if file_obj:
                paint_proxy_badge(painter, deco_rect, proxy_service.get_proxy_state(file_obj))

        self._paint_progress_bar(painter, deco_rect, generation_badge, QColor("#53A0ED"), 0)
        self._paint_progress_bar(painter, deco_rect, proxy_badge, QColor("#3AA1FF"), 1)

    def _paint_progress_bar(self, painter, deco_rect, badge, fill_color, stack_index):
        if not badge:
            return

        progress = int(badge.get("progress", 0))
        status = str(badge.get("status", "")).strip().lower()
        if status in ("queued", "running", "canceling"):
            progress = max(progress, 2)
        if progress <= 0:
            return

        bar_height = 3
        bar_margin = 2
        bottom_offset = (bar_height + 1) * int(stack_index)
        full_rect = deco_rect.adjusted(1, 0, -1, 0)
        full_rect.setTop(deco_rect.bottom() - bar_height - bar_margin - bottom_offset + 1)
        full_rect.setHeight(bar_height)
        if full_rect.width() <= 2:
            return

        fill_width = max(1, int((full_rect.width() * min(progress, 100)) / 100.0))
        fill_rect = full_rect.adjusted(0, 0, -(full_rect.width() - fill_width), 0)

        painter.save()
        painter.setPen(Qt.NoPen)
        painter.setBrush(QColor("#283241"))
        painter.drawRect(full_rect)
        painter.setBrush(fill_color)
        painter.drawRect(fill_rect)
        if status == "queued":
            label = "Queued"
            fm = QFontMetrics(painter.font())
            text_w = font_metrics_horizontal_advance(fm, label)
            text_h = fm.height()
            pad_x = 5
            pad_y = 2
            badge_w = text_w + (pad_x * 2)
            badge_h = text_h + (pad_y * 2)
            badge_bottom = full_rect.top() - 3
            badge_top = max(deco_rect.top() + 3, badge_bottom - badge_h + 1)
            badge_rect = deco_rect.adjusted(3, badge_top - deco_rect.top(), 0, 0)
            badge_rect.setWidth(badge_w)
            badge_rect.setHeight(badge_h)
            painter.setBrush(QColor(18, 22, 30, 220))
            painter.drawRoundedRect(badge_rect, 4, 4)
            painter.setPen(QColor("#EAF5FF"))
            painter.drawText(badge_rect, Qt.AlignCenter, label)
        painter.restore()


class FilesTreeView(QTreeView):
    """ A TreeView QWidget used on the main window """
    drag_item_size = QSize(48, 48)
    drag_item_center = QPoint(24, 24)

    def contextMenuEvent(self, event):

        # Set context menu mode
        app = get_app()
        _ = app._tr
        app.context_menu_object = "files"

        event.accept()
        index = self.indexAt(event.pos())
        if not index.isValid():
            self.clearSelection()
        else:
            self.selectionModel().setCurrentIndex(index, QItemSelectionModel.NoUpdate)

        # Build menu
        menu = StyledContextMenu(parent=self)

        menu.addAction(self.win.actionImportFiles)

        source_file = None
        active_job = None
        file_id = None
        if index.isValid():
            id_index = index.sibling(index.row(), 5)
            file_id = index.model().data(id_index, Qt.DisplayRole)
            if _is_generation_placeholder(file_id):
                job_id = _job_id_from_placeholder(file_id)
                queue = getattr(self.win, "generation_queue", None)
                active_job = queue.get_job(job_id) if queue else None
                if active_job and active_job.get("status") not in ("queued", "running", "canceling"):
                    active_job = None
            else:
                active_job = self.win.active_generation_job_for_file(file_id)
                source_file = File.get(id=file_id)
        add_ai_tools_menu(self.win, menu, source_file=source_file)

        if not active_job:
            self.win.actionGenerate.setEnabled(self.win.can_open_generate_dialog())
        if active_job:
            cancel_action = menu.addAction(_("Cancel Job"))
            delete_icon_path = os.path.join(info.PATH, "themes", "cosmic", "images", "track-delete-enabled.svg")
            if os.path.exists(delete_icon_path):
                cancel_action.setIcon(QIcon(delete_icon_path))
            else:
                cancel_action.setIcon(self.win.actionRemove_from_Project.icon())
            cancel_action.triggered.connect(
                lambda checked=False, job_id=active_job.get("id"): self.win.cancel_generation_job(job_id)
            )
        menu.addSeparator()
        menu.addAction(self.win.actionThumbnailView)

        if index.isValid():
            # Look up the model item and our unique ID
            model = index.model()

            # Look up file_id from 5th column of row
            id_index = index.sibling(index.row(), 5)
            file_id = model.data(id_index, Qt.DisplayRole)

            # If a valid file selected, show file related options
            menu.addSeparator()

            # Add edit title option (if svg file)
            file = File.get(id=file_id)
            if not file:
                menu.popup(event.globalPos())
                return
            if file and file.data.get("path").endswith(".svg"):
                menu.addAction(self.win.actionEditTitle)
                menu.addAction(self.win.actionDuplicate)
                menu.addSeparator()

            menu.addAction(self.win.actionPreview_File)
            add_optimized_preview_menu(self.win, menu)
            menu.addSeparator()
            menu.addAction(self.win.actionSplitFile)
            menu.addAction(self.win.actionExportFiles)
            menu.addSeparator()
            menu.addAction(self.win.actionAdd_to_Timeline)

            # Add Profile menu
            profile_menu = StyledContextMenu(title=_("Choose Profile"), parent=self)
            profile_icon = get_app().window.actionProfile.icon()
            profile_missing_icon = QIcon(":/icons/Humanity/actions/16/list-add.svg")
            profile_menu.setIcon(profile_icon)

            # Get file's profile
            file_profile = file.profile()
            if file_profile.info.description:
                action = profile_menu.addAction(profile_icon, file_profile.info.description)
                action.triggered.connect(lambda: get_app().window.actionProfile_trigger(file_profile))
            else:
                action = profile_menu.addAction(profile_missing_icon, _(f"Create Profile") + f": {file_profile.ShortName()}")
                action.triggered.connect(lambda: get_app().window.actionProfileEdit_trigger(file_profile))
            menu.addMenu(profile_menu)

            menu.addAction(self.win.actionFile_Properties)
            menu.addSeparator()
            menu.addAction(self.win.actionRemove_from_Project)
            menu.addSeparator()

        # Show menu
        menu.popup(event.globalPos())

    def mousePressEvent(self, event):
        index = self.indexAt(event.pos())
        if not index.isValid() and event.button() in (Qt.LeftButton, Qt.RightButton):
            self.clearSelection()
        super().mousePressEvent(event)

    def mouseDoubleClickEvent(self, event):
        # Get the index of the item at the click position
        index = self.indexAt(event.pos())
        if index.column() == 0:
            # If column 0 (thumbnail) is double-clicked, trigger the custom actions
            if int(get_app().keyboardModifiers() & Qt.ShiftModifier) > 0:
                get_app().window.actionSplitFile.trigger()
            elif int(get_app().keyboardModifiers() & Qt.ControlModifier) > 0:
                get_app().window.actionFile_Properties.trigger()
            else:
                get_app().window.actionPreview_File.trigger()
        else:
            super(FilesTreeView, self).mouseDoubleClickEvent(event)

    def dragEnterEvent(self, event):
        # If dragging urls onto widget, accept
        if event.mimeData().hasUrls():
            event.setDropAction(Qt.CopyAction)
            event.accept()

    def startDrag(self, supportedActions):
        """ Override startDrag method to display custom icon """

        # Get first column indexes for all selected rows
        selected = self.selectionModel().selectedRows(0)
        selected = [idx for idx in selected if not _is_generation_placeholder(idx.sibling(idx.row(), 5).data(Qt.DisplayRole))]

        # Check if there are any selected items
        if not selected:
            log.warning("No draggable items found in model!")
            return False

        # Get icons from up to 3 selected items
        icons = []
        for i in range(min(3, len(selected))):
            current = selected[i]
            icon = current.sibling(current.row(), 0).data(Qt.DecorationRole)
            if icon:
                icons.append(icon.pixmap(self.drag_item_size))

        # If no icons were retrieved, abort the drag
        if not icons:
            log.warning("No valid icons found for dragging!")
            return False

        # Calculate the total width of the composite pixmap including gaps
        gap = 1  # 1 pixel gap between icons
        total_width = (self.drag_item_size.width() * len(icons)) + (gap * (len(icons) - 1))

        # Create a composite pixmap to hold the icons in a row
        composite_pixmap = QPixmap(total_width, self.drag_item_size.height())
        composite_pixmap.fill(Qt.transparent)  # Start with a transparent background

        # Use a QPainter to draw the icons in a row with 1 pixel gap between them
        painter = QPainter(composite_pixmap)
        for idx, icon_pixmap in enumerate(icons):
            x_offset = idx * (self.drag_item_size.width() + gap)  # Position each icon with a gap
            painter.drawPixmap(int(x_offset), 0, icon_pixmap)
        painter.end()

        # Start the drag operation
        drag = QDrag(self)

        # Combine all selected items into the mime data
        mime_data = self.model().mimeData(selected)
        drag.setMimeData(mime_data)

        # Set the composite pixmap for the drag operation
        drag.setPixmap(composite_pixmap)

        # Set the hot spot to the center of the composite pixmap
        drag.setHotSpot(composite_pixmap.rect().center())

        # Start a transaction so all clips are grouped for a single undo
        tid = str(uuid.uuid4())
        get_app().updates.transaction_id = tid

        # Execute the drag operation (blocking - dropEvent creates clips during this call)
        drag.exec_(supportedActions)

        # End transaction
        get_app().updates.transaction_id = None

    # Without defining this method, the 'copy' action doesn't show with cursor
    def dragMoveEvent(self, event):
        event.accept()

    # Handle a drag and drop being dropped on widget
    def dropEvent(self, event):
        if not event.mimeData().hasUrls():
            # Nothing we're interested in
            event.ignore()
            return
        event.accept()
        # Use try/finally so we always reset the cursor
        try:
            # Set cursor to waiting
            get_app().setOverrideCursor(QCursor(Qt.WaitCursor))

            qurl_list = event.mimeData().urls()
            log.info("Processing drop event for {} urls".format(len(qurl_list)))
            self.files_model.process_urls(qurl_list)
        finally:
            # Restore cursor
            get_app().restoreOverrideCursor()

    # Forward file-add requests to the model, for legacy code (previous API)
    def add_file(self, filepath):
        self.files_model.add_files(filepath)

    def filter_changed(self):
        self.refresh_view()

    def refresh_view(self):
        """Resize and hide certain columns"""
        self.hideColumn(3)
        self.hideColumn(4)
        self.hideColumn(5)
        self.resize_contents()

    def resize_contents(self):
        # Get size of widget
        thumbnail_width = 80
        tags_width = 75

        # Resize thumbnail and tags column
        self.header().resizeSection(0, thumbnail_width)
        self.header().resizeSection(2, tags_width)

        # Set stretch mode on certain columns
        self.header().setStretchLastSection(False)
        self.header().setSectionResizeMode(1, QHeaderView.Stretch)
        self.header().setSectionResizeMode(2, QHeaderView.Interactive)

    def value_updated(self, item):
        """ Name or tags updated """
        if self.files_model.ignore_updates:
            return
        if item is None:
            return
        if item.column() not in (1, 2):
            return

        # Determine what was changed
        file_id_item = self.files_model.model.item(item.row(), 5)
        name_item = self.files_model.model.item(item.row(), 1)
        tags_item = self.files_model.model.item(item.row(), 2)
        if not file_id_item or not name_item or not tags_item:
            return

        file_id = file_id_item.text()
        if _is_generation_placeholder(file_id):
            return

        name = name_item.text()
        tags = tags_item.text()

        # Get file object and update friendly name and tags attribute
        f = File.get(id=file_id)
        if not f:
            return
        f.data.update({"name": name or os.path.basename(f.data.get("path"))})
        if "tags" in f.data or tags:
            f.data.update({"tags": tags})

        # Save File
        f.save()

        # Update file thumbnail
        self.win.FileUpdated.emit(file_id)

    def __init__(self, model, *args):
        # Invoke parent init
        super().__init__(*args)

        # Get a reference to the window object
        self.win = get_app().window

        # Get Model data
        self.files_model = model
        self.setModel(self.files_model.proxy_model)

        # Remove the default selection model and wire up to the shared one
        self.selectionModel().deleteLater()
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionModel(self.files_model.selection_model)
        self.setSortingEnabled(True)
        self.setItemDelegate(FilesTreeProgressDelegate(self))

        self.setAcceptDrops(True)
        self.setDragEnabled(True)
        self.setDropIndicatorShown(True)

        # Setup header columns and layout
        self.setIconSize(info.TREE_ICON_SIZE)
        self.setIndentation(0)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.setWordWrap(False)
        self.setTextElideMode(Qt.ElideRight)

        self.files_model.ModelRefreshed.connect(self.refresh_view)

        # setup filter events
        self.files_model.model.itemChanged.connect(self.value_updated)



"""
 @file
 @brief This file contains some Effect metadata related to pre-processing of Effects
 @author Jonathan Thomas <jonathan@openshot.org>
 @author Frank Dana <ferdnyc AT gmail com>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """
from classes.info import YOLO_PATH
import os
# Not all Effects support pre-processing, so for now, this is a hard-coded
# solution to providing the pre-processing params needed for these special effects.

effect_options = {
    # TODO: Remove Example example options
    "Example": [
        {
            "title": "Region",
            "setting": "region",
            "x": 0.05,
            "y": 0.05,
            "width": 0.25,
            "height": 0.25,
            "type": "rect"
        },
        {
            "max": 100,
            "title": "Volume",
            "min": 0,
            "setting": "volume",
            "value": 75,
            "type": "spinner"
        },
        {
            "value": "blender",
            "title": "Blender Command (path)",
            "type": "text",
            "setting": "blender_command"
        },
        {
            "max": 192000,
            "title": "Default Audio Sample Rate",
            "min": 22050,
            "setting": "default-samplerate",
            "value": 48000,
            "values": [
                {
                    "value": 22050,
                    "name": "22050"
                },
                {
                    "value": 44100,
                    "name": "44100"
                },
                {
                    "value": 48000,
                    "name": "48000"
                },
                {
                    "value": 96000,
                    "name": "96000"
                },
                {
                    "value": 192000,
                    "name": "192000"
                }
            ],
            "type": "dropdown",
        }
    ],

    "Tracker": [
        {
            "title": "Region",
            "setting": "region",
            "x": 0.05,
            "y": 0.05,
            "width": 0.25,
            "height": 0.25,
            "first-frame": 1,
            "type": "rect"
        },

        {
            "title": "Tracker type",
            "setting": "tracker-type",
            "value": "KCF",
            "values": [
                {
                    "value": "KCF",
                    "name": "KCF"
                },
                {
                    "value": "MIL",
                    "name": "MIL"
                },
                {
                    "value": "BOOSTING",
                    "name": "BOOSTING"
                },
                {
                    "value": "TLD",
                    "name": "TLD"
                },
                {
                    "value": "MEDIANFLOW",
                    "name": "MEDIANFLOW"
                },
                {
                    "value": "MOSSE",
                    "name": "MOSSE"
                },
                {
                    "value": "CSRT",
                    "name": "CSRT"
                }
            ],
            "type": "dropdown",
        }
    ],

    "Stabilizer": [
        {
            "max": 100,
            "title": "Smoothing window",
            "min": 1,
            "setting": "smoothing-window",
            "value": 30,
            "type": "spinner"
        }
    ],

    "ObjectDetection": [
        {
            "value": "https://github.com/OpenShot/openshot-qt/wiki/Object-Detection-Effect-(dependencies)",
            "title": "Click here for instructions and dependencies...",
            "type": "link",
            "setting": "model-instructions"
        },
        {
            "value": os.path.join(YOLO_PATH,"yolov3.weights"),
            "title": "Model Weights",
            "type": "text",
            "setting": "model-weights"
        },
        {
            "value": os.path.join(YOLO_PATH,"yolov3.cfg"),
            "title": "Model Config",
            "type": "text",
            "setting": "model-config"
        },
        {
            "value": os.path.join(YOLO_PATH,"obj.names"),
            "title": "Class names",
            "type": "text",
            "setting": "class-names"
        },
        {
            "title": "Processing Device",
            "setting": "processing-device",
            "value": "GPU",
            "values": [
                {
                    "value": "GPU",
                    "name": "GPU"
                },
                {
                    "value": "CPU",
                    "name": "CPU"
                }
            ],
            "type": "dropdown",
        }
    ]
}



"""
 @file
 @brief This file loads the animated title dialog (i.e Blender animation automation)
 @author Noah Figg <eggmunkee@hotmail.com>
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import uuid

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QApplication, QDialog, QDialogButtonBox, QPushButton
)

from classes import info, ui_util, metrics, tabstops
from classes.app import get_app
from classes.logger import log
from windows.views.blender_listview import BlenderListView


class AnimatedTitle(QDialog):
    """ Animated Title Dialog """

    # Path to ui file
    ui_path = os.path.join(info.PATH, 'windows', 'ui', 'animated-title.ui')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Load UI from designer & init
        ui_util.load_ui(self, self.ui_path)
        ui_util.init_ui(self)

        metrics.track_metric_screen("animated-title-screen")

        app = get_app()
        _ = app._tr

        # Add render controls
        self.btnRender = QPushButton(_('Render'))
        self.btnRender.setObjectName("acceptButton")
        self.btnCancel = QPushButton(_('Cancel'))
        self.btnCancel.setObjectName("cancelButton")
        self.buttonBox.addButton(self.btnRender, QDialogButtonBox.AcceptRole)
        self.buttonBox.addButton(self.btnCancel, QDialogButtonBox.RejectRole)
        # Set focus policy after adding to buttonBox to prevent override
        self.btnRender.setFocusPolicy(Qt.StrongFocus)
        self.btnCancel.setFocusPolicy(Qt.StrongFocus)

        # Hide render progress until needed
        self.statusContainer.hide()

        # Add blender view
        self.blenderView = BlenderListView(self)
        self.verticalLayout.addWidget(self.blenderView)

        self.imgPreview.setFocusPolicy(Qt.NoFocus)
        self.scrollArea.setFocusPolicy(Qt.NoFocus)
        self.buttonBox.setFocusPolicy(Qt.NoFocus)
        self.splitter.setFocusPolicy(Qt.NoFocus)

        # Init variables
        self.unique_folder_name = str(uuid.uuid1())
        self.output_dir = os.path.join(info.USER_PATH, "blender")
        self.selected_template = ""
        self.is_rendering = False
        self.my_blender = None

        # Clear all child controls
        self.clear_effect_controls()

        self._apply_tab_order()

    def _apply_tab_order(self):
        """Apply explicit tab order for animated title dialog."""
        from PyQt5.QtWidgets import QWidget
        from PyQt5.QtCore import QTimer

        def do_tab_order():
            # Force focus policies on buttons (something is resetting Cancel to NoFocus)
            self.btnCancel.setFocusPolicy(Qt.StrongFocus)
            self.btnRender.setFocusPolicy(Qt.StrongFocus)

            ordered = []
            if getattr(self, "blenderView", None):
                ordered.append(self.blenderView)
            ordered.extend([self.sliderPreview, self.btnRefresh])
            ordered.extend(
                tabstops.collect_focusable_from_layout(
                    self.settingsContainer.layout(),
                    self,
                    include_hidden=True,
                    include_disabled=True,
                )
            )
            ordered.extend([self.btnCancel, self.btnRender])

            # Apply tab order directly
            for first, second in zip(ordered, ordered[1:]):
                QWidget.setTabOrder(first, second)
            if len(ordered) >= 2:
                QWidget.setTabOrder(ordered[-1], ordered[0])

        QTimer.singleShot(0, do_tab_order)

    def accept(self):
        """ Start rendering animation, but don't close window """
        # Render
        self.blenderView.Render()

    def closeEvent(self, event):
        """ Actually close window and accept dialog """
        self.blenderView.Cancel()
        self.blenderView.end_processing()
        super().accept()

    def reject(self):
        # Stop threads
        self.blenderView.Cancel()
        self.blenderView.end_processing()
        super().reject()

    def clear_effect_controls(self):
        """ Clear all child widgets used for settings """
        self.statusContainer.hide()

        # Loop through child widgets
        for child in self.settingsContainer.children():
            try:
                self.settingsContainer.layout().removeWidget(child)
                child.deleteLater()
            except Exception:
                log.debug('Failed to remove child widget for effect controls')



"""
 @file
 @brief This file contains the clip properties model, used by the properties view
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
from collections import OrderedDict
from operator import itemgetter

from PyQt5.QtCore import QMimeData, Qt, QLocale, QTimer
from PyQt5.QtGui import (
    QStandardItemModel, QStandardItem,
    QPixmap, QColor,
    )

from classes.waveform import get_audio_data
from classes import info, updates
from classes import openshot_rc  # noqa
from classes.clip_utils import clamp_timing_to_media, clip_time_bounds, is_single_image_media
from classes.query import Clip, Transition, Effect, File
from classes.logger import log
from classes.app import get_app
import openshot

import json


class ClipStandardItemModel(QStandardItemModel):
    def __init__(self, parent=None):
        QStandardItemModel.__init__(self)

    def mimeData(self, indexes):
        # Create MimeData for drag operation
        data = QMimeData()

        # Get list of all selected file ids
        property_names = []
        for item in indexes:
            selected_row = self.itemFromIndex(item).row()
            property_names.append(self.item(selected_row, 0).data())
        data.setText(json.dumps(property_names))

        # Return Mimedata
        return data


class PropertiesModel(updates.UpdateInterface):
    def _resolve_reader_source_path(self, value):
        """Resolve a reader property value to a filesystem path when possible."""
        if value in (None, ""):
            return ""

        if isinstance(value, dict):
            value = value.get("path", "")

        source_value = str(value).strip()
        if not source_value:
            return ""

        if os.path.exists(source_value):
            return source_value

        file_obj = File.get(path=source_value)
        if file_obj:
            return file_obj.absolute_path()

        return source_value

    def _reader_display_name(self, selected_item, reader_memo):
        """Resolve a human-friendly reader display name from File.name when available."""
        reader_json = json.loads(reader_memo or "{}")
        reader_path = reader_json.get("path", "")
        file_id = reader_json.get("id")

        if getattr(selected_item, "data", None):
            file_id = selected_item.data.get("file_id") or file_id

        if file_id:
            file_obj = File.get(id=file_id)
            if file_obj:
                file_name = str(file_obj.data.get("name", "")).strip()
                if file_name:
                    return file_name

        return os.path.basename(reader_path)

    def _mask_reader_data(self, data):
        reader = {}
        if isinstance(data, dict):
            for key in ("mask_reader", "reader"):
                candidate = data.get(key)
                if isinstance(candidate, dict):
                    reader = candidate
                    break
        return reader

    def _transition_uses_static_mask(self, transition_data):
        reader = self._mask_reader_data(transition_data)
        if not reader:
            return True
        if not reader.get("path") and not reader.get("id") and not reader.get("has_single_image"):
            return True
        if isinstance(reader, dict) and "has_single_image" in reader:
            return bool(reader.get("has_single_image"))
        return bool(is_single_image_media(reader))

    def _mask_source_is_animated(self, data):
        reader = self._mask_reader_data(data)
        return bool(isinstance(reader, dict) and reader.get("has_single_image") is False)

    def _transition_default_keyframes(self, duration, start_value, end_value, contrast_value):
        fps = get_app().project.get("fps")
        fps_float = float(fps["num"]) / float(fps["den"])
        duration = max(0.0, float(duration or 0.0))

        brightness = openshot.Keyframe()
        brightness.AddPoint(1, float(start_value), openshot.BEZIER)
        if float(start_value) != float(end_value):
            brightness.AddPoint(round(duration * fps_float) + 1, float(end_value), openshot.BEZIER)
        contrast = openshot.Keyframe(float(contrast_value))
        return json.loads(brightness.Json()), json.loads(contrast.Json())

    def _effect_filter_base_properties(self, effect_data):
        """Return effect base properties that should remain hidden."""
        hidden = ["position", "layer", "start", "end", "duration"]
        if not isinstance(effect_data, dict):
            return hidden

        is_mask = (
            effect_data.get("type") == "Mask"
            or effect_data.get("class_name") == "Mask"
            or "mask_reader" in effect_data
        )
        if is_mask and self._mask_source_is_animated(effect_data):
            return ["position", "layer", "duration"]
        return hidden

    def _refresh_selected_effect_filters(self):
        if not self.selected:
            return
        item, item_type = self.selected[0]
        if item_type != "effect":
            return

        effect_id = None
        try:
            effect_id = item.Id()
        except Exception:
            effect_id = None
        effect_obj = Effect.get(id=effect_id) if effect_id else None
        next_filter_base_properties = self._effect_filter_base_properties(
            getattr(effect_obj, "data", None)
        )
        if next_filter_base_properties != self.filter_base_properties:
            self.filter_base_properties = next_filter_base_properties
            self.new_item = True
        else:
            self.filter_base_properties = next_filter_base_properties

    def _resolve_tracked_object_id(self, raw_properties, tracked_objects_raw_properties):
        """Resolve selected tracked object ID from properties payload."""
        if not tracked_objects_raw_properties:
            return None

        selected_idx = raw_properties.get("selected_object_index", {}).get("value")
        if selected_idx not in (None, "", "None"):
            selected_idx = str(selected_idx)
            if selected_idx in tracked_objects_raw_properties:
                return selected_idx

            # Newer tracked-object IDs are "<effect-uuid>-<index>".
            suffix = f"-{selected_idx}"
            for object_id in tracked_objects_raw_properties.keys():
                if object_id.endswith(suffix):
                    return object_id

        # Fallback to first tracked object
        return next(iter(tracked_objects_raw_properties.keys()))

    # This method is invoked by the UpdateManager each time a change happens (i.e UpdateInterface)
    def changed(self, action):

        # Handle change
        if action and len(action.key) >= 1 and action.key[0] in ["clips", "effects"] and action.type in ["update", "insert"]:
            log.debug(action.values)
            self._refresh_selected_effect_filters()
            # Update the model data
            self.update_model(get_app().window.txtPropertyFilter.text())

    # Update the selected item (which drives what properties show up)
    def update_item(self, selection):
        """Update selected items list"""
        self.next_selection = selection

        if not selection and self.selected:
            self.update_item_timeout()
        else:
            self.update_timer.start()

    # Update the next item (once the timer runs out)
    def update_item_timeout(self):
        # Get the next item id, and type
        selection = self.next_selection or []

        # Clear previous selection
        self.selected = []
        self.filter_base_properties = []

        if selection:
            timeline = get_app().window.timeline_sync.timeline
            for sel in selection:
                item_id = sel.get("id")
                item_type = sel.get("type")
                if item_type == "clip":
                    c = timeline.GetClip(item_id)
                    if c:
                        self.selected.append((c, item_type))
                elif item_type == "transition":
                    t = timeline.GetEffect(item_id)
                    if t:
                        self.selected.append((t, item_type))
                elif item_type == "effect":
                    e = timeline.GetClipEffect(item_id)
                    if e:
                        self.selected.append((e, item_type))
                        self.selected_parent = e.ParentClip()
                        self._refresh_selected_effect_filters()

            # Update frame # from timeline
            self.update_frame(get_app().window.preview_thread.player.Position(), reload_model=False)

            # Get ID of item
            self.new_item = True

        # Update the model data
        self.update_model(get_app().window.txtPropertyFilter.text())

    # Update the values of the selected clip, based on the current frame
    def update_frame(self, frame_number, reload_model=True):

        # Check for a selected clip
        if self.selected:
            clip, item_type = self.selected[0]

            if not clip:
                # Ignore null clip
                return

            # If effect, we really care about the position of the parent clip
            if item_type == "effect":
                clip = self.selected_parent

            # Get FPS from project
            fps = get_app().project.get("fps")
            fps_float = float(fps["num"]) / float(fps["den"])

            # Requested time
            requested_time = float(frame_number - 1) / fps_float

            # Determine the frame needed for this clip (based on the position on the timeline)
            time_diff = (requested_time - clip.Position()) + clip.Start()
            new_frame_number = round(time_diff * fps_float) + 1
            if new_frame_number != self.frame_number:
                # If frame # has changed
                self.frame_number = new_frame_number

                # Calculate biggest and smallest possible frames
                min_frame_number = round((clip.Start() * fps_float)) + 1
                max_frame_number = round((clip.End() * fps_float))

                # Adjust frame number if out of range
                if self.frame_number < min_frame_number:
                    self.frame_number = min_frame_number
                if self.frame_number > max_frame_number:
                    self.frame_number = max_frame_number

                # Update the model data
                if reload_model:
                    self.update_model(get_app().window.txtPropertyFilter.text())

    def remove_keyframe(self, item):
        """Remove an existing keyframe (if any)"""

        # Determine what was changed
        property = self.model.item(item.row(), 0).data()
        property_type = property[1]["type"]
        closest_point_x = property[1]["closest_point_x"]
        property_type = property[1]["type"]
        property_key = property[0]
        object_id = property[1]["object_id"]
        item_data = item.data()

        for item_id, item_type in item_data:
            # Find this clip
            c = None
            clip_updated = False

            if item_type == "clip":
                # Get clip object
                c = Clip.get(id=item_id)
            elif item_type == "transition":
                # Get transition object
                c = Transition.get(id=item_id)
            elif item_type == "effect":
                # Get effect object
                c = Effect.get(id=item_id)

            if not c:
                return

            # Create reference
            clip_data = c.data
            if object_id:
                objects = c.data.get('objects', {})
                clip_data = objects.pop(object_id, {})
                if not clip_data:
                    log.debug("No clip data found for this object id")
                    return

            if property_key in clip_data:  # Update clip attribute
                log_id = "{}/{}".format(item_id, object_id) if object_id else item_id
                log.debug("%s: remove %s keyframe. %s", log_id, property_key, clip_data.get(property_key))

                # Determine type of keyframe (normal or color)
                keyframe_list = []
                if property_type == "color":
                    keyframe_list = [clip_data[property_key]["red"], clip_data[property_key]["blue"], clip_data[property_key]["green"]]
                else:
                    keyframe_list = [clip_data[property_key]]

                # Loop through each keyframe (red, blue, and green)
                for keyframe_index, keyframe in enumerate(keyframe_list):
                    # Keyframe
                    # Loop through points, find a matching points on this frame
                    closest_point = None
                    point_to_delete = None
                    for point in keyframe["Points"]:
                        if point["co"]["X"] == self.frame_number:
                            # Found point, Update value
                            clip_updated = True
                            point_to_delete = point
                            break
                        if point["co"]["X"] == closest_point_x:
                            closest_point = point

                    # If no point found, use closest point x
                    if not point_to_delete:
                        point_to_delete = closest_point

                    # Delete point (if needed)
                    if point_to_delete:
                        clip_updated = True
                        log.debug("Found point to delete at X=%s" % point_to_delete["co"]["X"])
                        keyframe["Points"].remove(point_to_delete)

                        # Check for 0 keyframes (and use sane defaults instead of the 0.0 default value)
                        default_value = None
                        if not keyframe["Points"]:
                            if property_key in ["alpha", "scale_x", "scale_y", "time", "volume"]:
                                default_value = 1.0
                            elif property_key in ["origin_x", "origin_y"]:
                                default_value = 0.5
                            elif property_key in ["location_x", "location_y", "rotation", "shear_x", "shear_y"]:
                                default_value = 0.0
                            elif property_key in ["has_audio", "has_video", "channel_filter", "channel_mapping"]:
                                default_value = -1.0
                            elif property_key in ["wave_color"]:
                                if keyframe_index == 0:
                                    # Red
                                    default_value = 0.0
                                elif keyframe_index == 1:
                                    # Blue
                                    default_value = 255.0
                                elif keyframe_index == 2:
                                    # Green
                                    default_value = 123.0
                            if default_value is not None:
                                keyframe["Points"].append({
                                    'co': {'X': self.frame_number, 'Y': default_value},
                                    'interpolation': 1})

                # Enforce clip timing constraints
                if property_key in ("time", "start", "end", "duration"):
                    clamp_timing_to_media(clip_data, c)

                # Determine if waveforms are impacted by this change
                has_waveform = False
                waveform_file_id = None
                if property_key == "volume":
                    if clip_data.get("ui", {}).get("audio_data", []):
                        waveform_file_id = c.data.get("file_id")
                        has_waveform = True

                use_transition_update = item_type == "transition" and not object_id and property_key == "mask_reader"
                full_transition_data = clip_data if use_transition_update else None

                # Reduce # of clip properties we are saving (performance boost)
                if use_transition_update:
                    pass
                elif not object_id:
                    if property_key == "time":
                        clip_data = {k: clip_data.get(k) for k in ("time", "end", "duration", "start")}
                    else:
                        clip_data = {property_key: clip_data.get(property_key)}
                else:
                    # If objects dict detected - don't reduce the # of objects
                    objects[object_id] = clip_data
                    clip_data = {'objects': objects}

                # Save changes
                if clip_updated:
                    if use_transition_update:
                        timeline = getattr(get_app().window, "timeline", None)
                        if timeline:
                            timeline.update_transition_data(full_transition_data, only_basic_props=False)
                            c = Transition.get(id=item_id) or c
                        else:
                            log.warning(
                                "Timeline unavailable for transition mask_reader update; saving transition %s directly",
                                item_id,
                            )
                            c.data = full_transition_data
                            c.save()
                    else:
                        # Save
                        c.data = clip_data
                        c.save()

                    # Update waveforms (if needed)
                    if has_waveform:
                        get_audio_data({waveform_file_id: [c.id]})

                    # Update the preview
                    if not self._trim_preview_mode:
                        get_app().window.refreshFrameSignal.emit()

                # Clear selection and restore focus to label column
                current_row = self.parent.currentIndex().row()
                self.parent.clearSelection()
                if current_row >= 0:
                    self.parent.setCurrentIndex(self.model.index(current_row, 0))

    def color_update(self, item, new_color, interpolation=-1, interpolation_details=[]):
        """Insert/Update a color keyframe for the selected row"""

        # Determine what was changed
        property = self.model.item(item.row(), 0).data()
        property_type = property[1]["type"]
        closest_point_x = property[1]["closest_point_x"]
        previous_point_x = property[1]["previous_point_x"]
        property_key = property[0]
        object_id = property[1]["object_id"]
        item_data = item.data()

        for item_id, item_type in item_data:
            if property_type == "color":
                # Find this clip
                c = None
                clip_updated = False

                if item_type == "clip":
                    # Get clip object
                    c = Clip.get(id=item_id)
                elif item_type == "transition":
                    # Get transition object
                    c = Transition.get(id=item_id)
                elif item_type == "effect":
                    # Get effect object
                    c = Effect.get(id=item_id)

                if c:
                    # Create reference
                    clip_data = c.data
                    if object_id:
                        objects = c.data.get('objects', {})
                        clip_data = objects.pop(object_id, {})
                        if not clip_data:
                            log.debug("No clip data found for this object id")
                            return

                    # Update clip attribute
                    if property_key in clip_data:
                        log_id = "{}/{}".format(item_id, object_id) if object_id else item_id
                        log.debug("%s: update color property %s. %s", log_id, property_key, clip_data.get(property_key))

                        # Loop through each keyframe (red, blue, green, and alpha)
                        for color, new_value in [
                                ("red", new_color.red()),
                                ("blue", new_color.blue()),
                                ("green", new_color.green()),
                                ("alpha", new_color.alpha()),
                                ]:

                            # Keyframe
                            # Make sure this color property exists in the clip data
                            if color not in clip_data[property_key]:
                                log.debug(f"Creating new color component: {color}")
                                clip_data[property_key][color] = {"Points": []}

                            # Loop through points, find a matching points on this frame
                            found_point = False
                            for point in clip_data[property_key][color].get("Points", []):
                                log.debug("looping points: co.X = %s" % point["co"]["X"])
                                if interpolation == -1 and point["co"]["X"] == self.frame_number:
                                    # Found point, Update value
                                    found_point = True
                                    clip_updated = True
                                    # Update point
                                    point["co"]["Y"] = new_value
                                    log.debug(
                                        "updating point: co.X = %d to value: %.3f",
                                        point["co"]["X"], float(new_value))
                                    break

                                elif interpolation > -1 and point["co"]["X"] == previous_point_x:
                                    # Only update the LEFT side of the curve (i.e. the previous point's right handle)
                                    found_point = True
                                    clip_updated = True
                                    if point.get("interpolation", 2) == 0 and interpolation_details:
                                        point["handle_right"] = point.get("handle_right") or {"Y": 0.0, "X": 0.0}
                                        point["handle_right"]["X"] = interpolation_details[0]
                                        point["handle_right"]["Y"] = interpolation_details[1]

                                        log.debug("updating previous point (right handle): co.X = %d", point["co"]["X"])
                                        log.debug("use interpolation preset: %s", str(interpolation_details))
                                    else:
                                        # Remove unused bezier property
                                        point.pop("handle_right", None)

                                elif interpolation > -1 and point["co"]["X"] == closest_point_x:
                                    # Only update interpolation type (and the RIGHT side of the curve)
                                    found_point = True
                                    clip_updated = True
                                    point["interpolation"] = interpolation
                                    if interpolation == 0 and interpolation_details:
                                        point["handle_left"] = point.get("handle_left") or {"Y": 0.0, "X": 0.0}
                                        point["handle_left"]["X"] = interpolation_details[2]
                                        point["handle_left"]["Y"] = interpolation_details[3]
                                    else:
                                        # Remove unused bezier property
                                        point.pop("handle_left", None)

                                    log.debug("updating interpolation mode point: co.X = %d to %d",
                                              point["co"]["X"], interpolation)
                                    log.debug("use interpolation preset: %s", str(interpolation_details))

                            # Create new point (if needed)
                            if not found_point:
                                clip_updated = True
                                log.debug("Created new point at X=%d", self.frame_number)
                                clip_data[property_key][color].setdefault("Points", []).append({
                                    'co': {'X': self.frame_number, 'Y': new_value},
                                    'interpolation': 1,
                                    })

                    # Reduce # of clip properties we are saving (performance boost)
                    if not object_id:
                        clip_data = {property_key: clip_data.get(property_key)}
                    else:
                        # If objects dict detected - don't reduce the # of objects
                        objects[object_id] = clip_data
                        clip_data = {'objects': objects}

                    # Save changes
                    if clip_updated:
                        # Save
                        c.data = clip_data
                        c.save()

                        # Update the preview
                        if not self._trim_preview_mode:
                            get_app().window.refreshFrameSignal.emit()

                    # Clear selection and restore focus to label column
                    current_row = self.parent.currentIndex().row()
                    self.parent.clearSelection()
                    if current_row >= 0:
                        self.parent.setCurrentIndex(self.model.index(current_row, 0))

    def value_updated(self, item, interpolation=-1, value=None, interpolation_details=[]):
        """ Table cell change event - also handles context menu to update interpolation value """

        if self.ignore_update_signal:
            return

        # Get translation method
        _ = get_app()._tr

        # Determine what was changed
        property = self.model.item(item.row(), 0).data()
        closest_point_x = property[1]["closest_point_x"]
        previous_point_x = property[1]["previous_point_x"]
        property_type = property[1]["type"]
        property_key = property[0]
        object_id = property[1]["object_id"]
        property_choices = property[1].get("choices") or []
        choice_keyframes_use_constant = bool(property_choices) and interpolation == -1
        objects = {}
        item_data = item.data()

        # Get value (if any)
        if item.text() or value:
            # Set and format value based on property type
            if value == "None":
                new_value = ""
            elif value is not None:
                # Override value
                new_value = value
            elif property_type == "string":
                # Use string value
                new_value = item.text()
            elif property_type == "bool":
                # Use boolean value
                if item.text() == _("False"):
                    new_value = False
                else:
                    new_value = True
            elif property_type == "int":
                # Use int value
                new_value = QLocale().system().toInt(item.text())[0]
            else:
                # Use decimal value
                new_value = QLocale().system().toFloat(item.text())[0]
        else:
            new_value = None

        for item_id, item_type in item_data:
            log.info(
                "%s for %s changed to %s at frame %s with interpolation: %s at closest x: %s",
                property_key, item_id, new_value, self.frame_number, interpolation, closest_point_x)

            # Start each iteration with the original value
            value = new_value

            # Find this clip
            c = None
            clip_updated = False

            if item_type == "clip":
                # Get clip object
                c = Clip.get(id=item_id)
            elif item_type == "transition":
                # Get transition object
                c = Transition.get(id=item_id)
            elif item_type == "effect":
                # Get effect object
                c = Effect.get(id=item_id)

            if c and c.data:
                is_mask_reader_update = (
                    not object_id
                    and property_key == "mask_reader"
                )

                # Create reference
                clip_data = c.data
                if object_id:
                    objects = c.data.get('objects', {})
                    clip_data = objects.pop(object_id, {})
                    if not clip_data:
                        log.debug("No clip data found for this object id")
                        return

                # Update clip attribute
                if property_key in clip_data:
                    log_id = "{}/{}".format(item_id, object_id) if object_id else item_id
                    log.debug("%s: update property %s. %s", log_id, property_key, clip_data.get(property_key))

                    # Check the type of property (some are keyframe, and some are not)
                    if property_type != "reader" and isinstance(clip_data[property_key], dict):
                        # Keyframe

                        # Protection from HUGE scale values
                        if property_key in ['scale_x', 'scale_y', 'shear_x', 'shear_y'] and value:
                            width = get_app().project.get("width")
                            height = get_app().project.get("height")
                            is_svg = clip_data.get("reader", {}).get("path", "").lower().endswith("svg")
                            if is_svg:
                                max_multiple = 15
                            else:
                                max_multiple = 50
                            if width > 0 and height > 0:
                                # Clamp the max scale based on project size
                                max_multiple = round((2000 * max_multiple) / max(width, height))

                            # Apply the calculated max_multiple to value
                            value = max(min(value, max_multiple), -max_multiple)

                        # Loop through points, find a matching points on this frame
                        found_point = False
                        point_to_delete = None
                        for point in clip_data[property_key].get('Points', []):
                            log.debug("looping points: co.X = %s" % point["co"]["X"])
                            if interpolation == -1 and point["co"]["X"] == self.frame_number:
                                # Found point, Update value
                                found_point = True
                                clip_updated = True
                                # Update or delete point
                                if value is not None:
                                    point["co"]["Y"] = int(value) if property_key == "time" else float(value)
                                    if choice_keyframes_use_constant:
                                        point["interpolation"] = openshot.CONSTANT
                                    log.debug("updating point: co.X = %d to value: %s",
                                              point["co"]["X"], value)
                                else:
                                    point_to_delete = point
                                break

                            if interpolation > -1 and point["co"]["X"] == previous_point_x:
                                # Only update the LEFT side of the curve (i.e. the previous point's right handle)
                                found_point = True
                                clip_updated = True
                                if point.get("interpolation", 2) == 0 and interpolation_details:
                                    point["handle_right"] = point.get("handle_right") or {"Y": 0.0, "X": 0.0}
                                    point["handle_right"]["X"] = interpolation_details[0]
                                    point["handle_right"]["Y"] = interpolation_details[1]

                                    log.debug("updating previous point (right handle): co.X = %d", point["co"]["X"])
                                    log.debug("use interpolation preset: %s", str(interpolation_details))
                                else:
                                    # Remove unused bezier property
                                    point.pop("handle_right", None)

                            if interpolation > -1 and point["co"]["X"] == closest_point_x:
                                # Only update the RIGHT side of the curve (i.e. the closest point's left handle
                                # and interpolation mode (only the RIGHT SIDE POINT controls the interpolation mode)
                                found_point = True
                                clip_updated = True
                                point["interpolation"] = interpolation
                                if interpolation == 0 and interpolation_details:
                                    point["handle_left"] = point.get("handle_left") or {"Y": 0.0, "X": 0.0}
                                    point["handle_left"]["X"] = interpolation_details[2]
                                    point["handle_left"]["Y"] = interpolation_details[3]
                                else:
                                    # Remove unused bezier property
                                    point.pop("handle_left", None)

                                log.debug("updating interpolation mode point: co.X = %d to %d",
                                          point["co"]["X"], interpolation)
                                log.debug("use interpolation preset: %s", str(interpolation_details))

                        # Delete point (if needed)
                        if point_to_delete:
                            clip_updated = True
                            log.debug("Found point to delete at X=%s" % point_to_delete["co"]["X"])
                            clip_data[property_key]["Points"].remove(point_to_delete)

                        # Create new point (if needed)
                        elif not found_point and value is not None:
                            clip_updated = True
                            log.debug("Created new point at X=%d", self.frame_number)
                            clip_data[property_key].setdefault('Points', []).append({
                                'co': {'X': self.frame_number, 'Y': int(value) if property_key == "time" else value},
                                'interpolation': openshot.CONSTANT if choice_keyframes_use_constant else openshot.LINEAR})

                if not clip_updated:
                    # If no keyframe was found, set a basic property
                    if property_type == "int":
                        clip_updated = True
                        try:
                            clip_data[property_key] = int(value)
                        except Exception:
                            log.warn('Invalid Integer value passed to property', exc_info=1)

                    elif property_type == "float":
                        clip_updated = True
                        try:
                            clip_data[property_key] = float(value)

                            # Fix precision issues with time properties by snapping to FPS grid
                            if property_key in ['position', 'start', 'end']:
                                fps_num = get_app().project.get("fps").get("num")
                                fps_den = get_app().project.get("fps").get("den")
                                frame_duration = fps_den / fps_num

                                # Snap the value to the nearest frame
                                clip_data[property_key] = round(clip_data[property_key] / frame_duration) * frame_duration

                        except Exception:
                            log.warn('Invalid Float value passed to property', exc_info=1)

                    elif property_type == "bool":
                        clip_updated = True
                        try:
                            clip_data[property_key] = bool(value)
                        except Exception:
                            log.warn('Invalid Boolean value passed to property', exc_info=1)

                    elif property_type == "string":
                        clip_updated = True
                        try:
                            clip_data[property_key] = str(value or "")
                        except Exception:
                            log.warn('Invalid String value passed to property', exc_info=1)

                    elif property_type in ["font", "caption"]:
                        clip_updated = True
                        try:
                            clip_data[property_key] = str(value)
                        except Exception:
                            log.warn('Invalid Font/Caption value passed to property', exc_info=1)

                    elif property_type == "reader":
                        # Reader / mask source
                        clip_updated = True
                        try:
                            selection_data = value if isinstance(value, dict) else {}
                            selected_start = selection_data.get("start")
                            selected_end = selection_data.get("end")
                            reader_value = {"type": ""}
                            if value:
                                # Set a new source
                                resolved_value = self._resolve_reader_source_path(value)
                                clip_object = openshot.Clip(resolved_value)
                                clip_object.Open()
                                reader_value = json.loads(clip_object.Reader().Json())
                                clip_object.Close()
                                clip_object = None
                            # Clear the source (set to a dict with a type field)
                            clip_data[property_key] = reader_value
                            if is_mask_reader_update:
                                clip_data["mask_reader"] = reader_value
                                clip_data["reader"] = reader_value

                            if is_mask_reader_update:
                                start_source = selected_start if selected_start is not None else clip_data.get("start", 0.0)
                                end_source = selected_end if selected_end is not None else clip_data.get("end", start_source)
                                start = float(start_source or 0.0)
                                end = float(end_source or start)
                                if end < start:
                                    end = start
                                duration = max(0.0, end - start)

                                if self._transition_uses_static_mask(clip_data):
                                    clip_data["start"] = 0.0
                                    clip_data["end"] = duration
                                    brightness, contrast = self._transition_default_keyframes(
                                        duration, 1.0, -1.0, 3.0
                                    )
                                else:
                                    clip_data["start"] = start
                                    clip_data["end"] = end
                                    brightness, contrast = self._transition_default_keyframes(
                                        duration, 0.0, 0.0, 0.0
                                    )

                                clip_data["duration"] = max(
                                    0.0,
                                    float(clip_data.get("end", 0.0) or 0.0) - float(clip_data.get("start", 0.0) or 0.0),
                                )
                                clip_data["brightness"] = brightness
                                clip_data["contrast"] = contrast
                        except Exception:
                            log.warn('Invalid Reader value passed to property: %s', value, exc_info=1)

                # Enforce clip timing constraints
                if property_key in ("time", "start", "end", "duration"):
                    clamp_timing_to_media(clip_data, c)

                # Determine if waveforms are impacted by this change
                has_waveform = False
                waveform_file_id = None
                if property_key == "volume":
                    if clip_data.get("ui", {}).get("audio_data", []):
                        waveform_file_id = c.data.get("file_id")
                        has_waveform = True

                # Reduce # of clip properties we are saving (performance boost)
                if is_mask_reader_update:
                    pass
                elif not object_id:
                    if property_key == "time":
                        clip_data = {k: clip_data.get(k) for k in ("time", "end", "duration", "start")}
                    else:
                        clip_data = {property_key: clip_data.get(property_key)}
                else:
                    # If objects dict detected - don't reduce the # of objects
                    objects[object_id] = clip_data
                    clip_data = {'objects': objects}

                # Save changes
                if clip_updated:
                    # Save
                    c.data = clip_data
                    c.save()

                    # Update waveforms (if needed)
                    if has_waveform:
                        get_audio_data({waveform_file_id: [c.id]})

                    # Update the preview
                    get_app().window.refreshFrameSignal.emit()

                    log.info("Item %s: changed %s to %s at frame %s (x: %s)" % (item_id, property_key, value, self.frame_number, closest_point_x))

                # Clear selection and restore focus to label column
                current_row = self.parent.currentIndex().row()
                self.parent.clearSelection()
                if current_row >= 0:
                    self.parent.setCurrentIndex(self.model.index(current_row, 0))

    def set_property(self, property, filter, c, item_type, object_id=None):
        app = get_app()
        _ = app._tr
        label = property[1]["name"]
        name = property[0]

        # Constrain time keyframes to the reader's frame range
        if name == "time" and getattr(c, "data", None):
            _, max_frames = clip_time_bounds(c.data, c)
            if max_frames:
                property[1]["min"] = 1
                property[1]["max"] = float(max_frames)

        value = property[1]["value"]
        type = property[1]["type"]
        memo = property[1]["memo"]
        readonly = property[1]["readonly"]
        keyframe = property[1]["keyframe"]
        points = property[1]["points"]
        interpolation = property[1]["interpolation"]
        choices = property[1]["choices"]
        # Add object id reference to QStandardItem
        property[1]["object_id"] = object_id

        # Adding Transparency to translation file
        transparency_label = _("Transparency")  # noqa

        selected_choice = None
        if choices:
            selected_choices = [c for c in choices if c.get("selected") is True]
            if selected_choices:
                selected_choice = selected_choices[0]["name"]

        # Hide filtered out properties
        if filter and filter.lower() not in _(label).lower():
            return

        # Hide unused base properties (if any)
        if name in self.filter_base_properties:
            return

        # Insert new data into model, or update existing values
        row = []
        if self.new_item:

            # Append Property Name
            col = QStandardItem("Property")
            col.setText(_(label))
            col.setData(property)
            if keyframe and points > 1:
                col.setBackground(QColor("green"))  # Highlight keyframe background
            elif points > 1:
                col.setBackground(QColor(42, 130, 218))  # Highlight interpolated value background
            if readonly or type in ["color", "font", "caption"] or choices or label == "Track":
                col.setFlags(Qt.ItemIsEnabled)
            else:
                col.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsUserCheckable)
            row.append(col)

            # Append Value
            col = QStandardItem("Value")
            if selected_choice:
                col.setText(_(selected_choice))
            elif type == "string":
                # Use string value
                col.setText(memo)
            elif type == "font":
                # Use font value
                col.setText(memo)
            elif type == "caption":
                # Use caption value
                col.setText(memo)
                # Load caption editor also
                get_app().window.CaptionTextLoaded.emit(memo, row)
            elif type == "bool":
                # Use boolean value
                if value:
                    col.setText(_("True"))
                else:
                    col.setText(_("False"))
            elif type == "color":
                # Don't output a value for colors
                col.setText("")
            elif type == "reader":
                col.setText(self._reader_display_name(c, memo))
            elif type == "int" and label == "Track":
                # Find track display name
                all_tracks = get_app().project.get("layers")
                display_count = len(all_tracks)
                display_label = None
                try:
                    value_int = int(float(value))
                except (TypeError, ValueError):
                    value_int = value
                for track in reversed(sorted(all_tracks, key=itemgetter('number'))):
                    if track.get("number") == value_int:
                        display_label = track.get("label")
                        break
                    display_count -= 1
                track_name = display_label or _("Track %s") % QLocale().toString(display_count)
                col.setText(track_name)

            elif type == "int":
                try:
                    int_value = int(float(value))
                    col.setText("%d" % int_value)
                except (TypeError, ValueError):
                    col.setText("" if value is None else str(value))
            else:
                # Use numeric value
                if value == "" or value is None:
                    col.setText("")
                else:
                    col.setText(QLocale().system().toString(float(value), "f", precision=3))
            col.setData([(obj.Id(), t) for obj, t in self.selected])
            if points > 1:
                # Apply icon to cell
                my_icon = QPixmap(":/curves/keyframe-%s.png" % interpolation)
                col.setData(my_icon, Qt.DecorationRole)

                # Set the background color of the cell
                if keyframe:
                    col.setBackground(QColor("green"))  # Highlight keyframe background
                else:
                    col.setBackground(QColor(42, 130, 218))  # Highlight interpolated value background

            if type == "color":
                # Color needs to be handled special
                vals = property[1]
                # r, g, b
                r, g, b = (int(vals[c]["value"]) for c in ("red", "green", "blue"))
                # alpha falls back to vals["max"] (usually 255) if missing
                a = int(vals.get("alpha", {}).get("value", vals.get("max", 255.0)))
                col.setBackground(QColor(r, g, b, a))

            if readonly or type in ["color", "font", "caption"] or choices or label == "Track":
                col.setFlags(Qt.ItemIsEnabled)
            else:
                col.setFlags(
                    Qt.ItemIsSelectable
                    | Qt.ItemIsEnabled
                    | Qt.ItemIsUserCheckable
                    | Qt.ItemIsEditable)
            row.append(col)

            # Append ROW to MODEL (if does not already exist in model)
            self.model.appendRow(row)

        elif name in self.items and self.items[name]["row"]:
            # Update the value of the existing model
            # Get 1st Column
            col = self.items[name]["row"][0]
            col.setData(property)

            # For non-color types, update the background color
            if keyframe and points > 1:
                col.setBackground(QColor("green"))  # Highlight keyframe background
            elif points > 1:
                col.setBackground(QColor(42, 130, 218))  # Highlight interpolated value background
            else:
                col.setBackground(QStandardItem("Empty").background())

            # Update helper dictionary
            row.append(col)

            # Get 2nd Column
            col = self.items[name]["row"][1]
            if selected_choice:
                col.setText(_(selected_choice))
            elif type == "string":
                # Use string value
                col.setText(memo)
            elif type == "font":
                # Use font value
                col.setText(memo)
            elif type == "caption":
                # Use caption value
                col.setText(memo)
            elif type == "bool":
                # Use boolean value
                if value:
                    col.setText(_("True"))
                else:
                    col.setText(_("False"))
            elif type == "color":
                # Don't output a value for colors
                col.setText("")
            elif type == "int" and label == "Track":
                # Find track display name
                all_tracks = get_app().project.get("layers")
                display_count = len(all_tracks)
                display_label = None
                try:
                    value_int = int(float(value))
                except (TypeError, ValueError):
                    value_int = value
                for track in reversed(sorted(all_tracks, key=itemgetter('number'))):
                    if track.get("number") == value_int:
                        display_label = track.get("label")
                        break
                    display_count -= 1
                track_name = display_label or _("Track %s") % QLocale().toString(display_count)
                col.setText(track_name)
            elif type == "int":
                try:
                    int_value = int(float(value))
                    col.setText("%d" % int_value)
                except (TypeError, ValueError):
                    col.setText("" if value is None else str(value))
            elif type == "reader":
                col.setText(self._reader_display_name(c, property[1].get("memo")))
            else:
                # Use numeric value
                if value == "" or value is None:
                    col.setText("")
                else:
                    col.setText(QLocale().system().toString(float(value), "f", precision=3))

            if points > 1:
                # Apply icon to cell
                my_icon = QPixmap(":/curves/keyframe-%s.png" % interpolation)
                col.setData(my_icon, Qt.DecorationRole)

                # Set the background color of the cell
                if keyframe:
                    col.setBackground(QColor("green"))  # Highlight keyframe background
                else:
                    col.setBackground(QColor(42, 130, 218))  # Highlight interpolated value background

            else:
                # clear background color
                col.setBackground(QStandardItem("Empty").background())

                # clear icon
                my_icon = QPixmap()
                col.setData(my_icon, Qt.DecorationRole)

            if type == "color":
                # Update the color based on the color curves
                vals = property[1]
                # r, g, b
                r, g, b = (int(vals[c]["value"]) for c in ("red", "green", "blue"))
                # alpha falls back to vals["max"] (usually 255) if missing
                a = int(vals.get("alpha", {}).get("value", vals.get("max", 255.0)))
                col.setBackground(QColor(r, g, b, a))

            # Update helper dictionary
            row.append(col)

        # Keep track of items in a dictionary (for quick look up)
        self.items[name] = {"row": row, "property": property}

    def update_model(self, filter=""):
        app = get_app()
        _ = app._tr

        if self.ignore_update_signal:
            log.debug("ignoring update signal, because we are already in an update...")
            return

        # Ignore any events from this method
        self.ignore_update_signal = True

        try:
            # Check for a selected clip
            if self.selected and self.selected[0]:
                c, item_type = self.selected[0]

                # Skip blank clips
                # TODO: Determine why c is occasional = None
                if not c:
                    return

                # Build list of raw properties for all selected items
                all_raw_properties = []
                for obj, _item_type in self.selected:
                    props = json.loads(obj.PropertiesJSON(self.frame_number))
                    all_raw_properties.append(props)

                # Use first item's properties as baseline
                raw_properties = all_raw_properties[0]
                tracked_object_id = None
                tracked_object_properties = {}

                if len(all_raw_properties) == 1:
                    tracked_objects_raw_properties = raw_properties.pop('objects', None)
                    if tracked_objects_raw_properties:
                        tracked_object_id = self._resolve_tracked_object_id(
                            raw_properties, tracked_objects_raw_properties)
                        tracked_object_properties = tracked_objects_raw_properties[tracked_object_id]
                        raw_properties.update(tracked_object_properties)
                else:
                    # Remove tracked object lists before comparing
                    if 'objects' in raw_properties:
                        raw_properties.pop('objects')
                    for props in all_raw_properties[1:]:
                        props.pop('objects', None)

                    # Determine shared properties across all selected items
                    shared = {}
                    for key, prop in raw_properties.items():
                        matches = True
                        for other in all_raw_properties[1:]:
                            if key not in other or other[key].get('type') != prop.get('type') or other[key].get('name') != prop.get('name'):
                                matches = False
                                break
                        if matches:
                            values = [other[key]['value'] for other in all_raw_properties]
                            memos = [other[key]['memo'] for other in all_raw_properties]
                            new_prop = prop.copy()
                            if all(v == values[0] for v in values):
                                new_prop['value'] = values[0]
                            else:
                                new_prop['value'] = ''
                            if all(m == memos[0] for m in memos):
                                new_prop['memo'] = memos[0]
                            else:
                                new_prop['memo'] = ''
                            shared[key] = new_prop

                    raw_properties = shared

                # Sort all properties (by 'name')
                all_properties = OrderedDict(sorted(raw_properties.items(), key=lambda x: x[1]['name']))

                # Check if filter was changed (if so, wipe previous model data)
                if self.previous_filter != filter:
                    self.previous_filter = filter
                    self.new_item = True  # filter changed, so we need to regenerate the entire model

                # Build or update the model
                if self.new_item:
                    # Prepare for new properties
                    self.items = {}
                    self.model.clear()

                    # Add Headers
                    self.model.setHorizontalHeaderLabels([_("Property"), _("Value")])

                    # Clear caption editor
                    get_app().window.CaptionTextLoaded.emit("", None)

                # Loop through properties, and build/update the model
                for property in all_properties.items():
                    if property[0] in tracked_object_properties:
                        # Add/update tracked object property
                        self.set_property(property, filter, c, item_type, object_id=tracked_object_id)
                    else:
                        # Add/update base property
                        self.set_property(property, filter, c, item_type)

                # After first render, future calls will update in place
                self.new_item = False

            else:
                # Clear previous model data (if any)
                self.model.clear()

                # Add Headers
                self.model.setHorizontalHeaderLabels([_("Property"), _("Value")])
        finally:
            # Done updating model (even if we returned early)
            self.ignore_update_signal = False

    def __init__(self, parent, *args):

        # Keep track of the selected items (clips, transitions, etc...)
        self.selected = []
        self.current_item_id = None
        self.frame_number = 1
        self.new_item = True
        self.items = {}
        self.ignore_update_signal = False
        self.parent = parent
        self.previous_filter = None
        self.filter_base_properties = []
        self._trim_preview_mode = False

        # Create standard model
        self.model = ClipStandardItemModel()
        self.model.setColumnCount(2)

        # Timer to use a delay before showing properties (to prevent a mass selection from trying
        # to update the property model hundreds of times)
        self.update_timer = QTimer(parent)
        self.update_timer.setInterval(100)
        self.update_timer.setSingleShot(True)
        self.update_timer.timeout.connect(self.update_item_timeout)
        self.next_selection = None

        # Connect data changed signal
        self.model.itemChanged.connect(self.value_updated)

        # Add self as listener to project data updates (used to update the timeline)
        get_app().updates.add_listener(self)
        get_app().window.TrimPreviewMode.connect(self._enter_trim_preview)
        get_app().window.TimelinePreviewMode.connect(self._exit_trim_preview)

    def _enter_trim_preview(self):
        self._trim_preview_mode = True

    def _exit_trim_preview(self):
        self._trim_preview_mode = False
